/*******************************************************************\

Copyright (c) 2008 Leiden University (LERC group at LIACS).
All rights reserved.

The use and distribution terms for this software are covered by the 
Common Public License 1.0 (http://opensource.org/licenses/cpl1.0.txt)
which can be found in the file LICENSE at the root of this distribution.
By using this software in any fashion, you are agreeing to be bound by 
the terms of this license.

You must not remove this notice, or any other, from this software.

\*******************************************************************/

#include <stdlib.h>
#include <string.h>
#include "DWT.h"

void mainDWT(const tile_c_t * tile_in, tile_c_t * tile_out)
{
	int i;

	tile_out->num = tile_in->num;
	tile_out->numcomp = tile_in->numcomp;

	// Copy the data from input to output structures
	for (i = 0; i < (WIDTH_TILE * HEIGHT_TILE); i++)
	{
		tile_out->comp[i] = tile_in->comp[i];
	}
	// Compute Wavelet for component
#ifdef INTEGER_TILES
		dwt_encode((int *) &tile_out->comp,tile_out->num);
#else 
		dwt_encode((char *) &tile_out->comp,tile_out->num);
#endif
}

#ifdef INTEGER_TILES
void dwt_encode(int * out, int num) {
	int i, j, k;
	int *a = NULL;
	int *aj = NULL;
	int *bj = NULL;
	int w, l;

	w = WIDTH_TILE;
	l = NUM_RES - 1;
	a = out;
	
	for (i = 0; i < l; i++) {
		int rw;			// width of the resolution level computed 
		int rh;			// heigth of the resolution level computed                                 
		int rw1;		// width of the resolution level once lower than computed one
		int rh1;		// height of the resolution level once lower than computed one
		int cas_col;	// 0 = non inversion on horizontal filtering 1 = inversion between low-pass and high-pass filtering
		int cas_row;	// 0 = non inversion on vertical filtering 1 = inversion between low-pass and high-pass filtering
		int dn, sn;
		int bj_alloc[WIDTH_TILE>HEIGHT_TILE?WIDTH_TILE:HEIGHT_TILE];
		
		rw = RES_WIDTH[l - i + RES_OFFSET];
		rh = RES_HEIGTH[l - i + RES_OFFSET];
		rw1= RES_WIDTH[l - i - 1 + RES_OFFSET];
		rh1= RES_HEIGTH[l - i - 1 + RES_OFFSET];
		
		int p = num % HTILES;
		int q = num / HTILES;
		
		cas_row = (p * RES_WIDTH[l - i + RES_OFFSET]) % 2;
		cas_col = (q * RES_HEIGTH[l - i + RES_OFFSET]) % 2;
        
		sn = rh1;
		dn = rh - rh1;
		bj = bj_alloc;
		for (j = 0; j < rw; j++) {
			aj = a + j;
			for (k = 0; k < rh; k++)  bj[k] = aj[k*w];
			dwt_encode_1(bj, dn, sn, cas_col);
			dwt_deinterleave_v(bj, aj, dn, sn, w, cas_col);
		}
		
		sn = rw1;
		dn = rw - rw1;
		bj = bj_alloc;
		for (j = 0; j < rh; j++) {
			aj = a + j * w;
			for (k = 0; k < rw; k++)  bj[k] = aj[k];
			dwt_encode_1(bj, dn, sn, cas_row);
			dwt_deinterleave_h(bj, aj, dn, sn, cas_row);
		}
	}
}

void dwt_encode_1(int *a, int dn, int sn, int cas) {
	int i;
	
	if (!cas) {
		if ((dn > 0) || (sn > 1)) {	
			for (i = 0; i < dn; i++) D(i) -= (S_(i) + S_(i + 1)) >> 1;
			for (i = 0; i < sn; i++) S(i) += (D_(i - 1) + D_(i) + 2) >> 2;
		}
	} else {
		if (!sn && dn == 1)
			S(0) *= 2;
		else {
			for (i = 0; i < dn; i++) S(i) -= (DD_(i) + DD_(i - 1)) >> 1;
			for (i = 0; i < sn; i++) D(i) += (SS_(i) + SS_(i + 1) + 2) >> 2;
		}
	}
}

void dwt_deinterleave_h(int *a, int *b, int dn, int sn, int cas) {
	int i;
    for (i=0; i<sn; i++) b[i]=a[2*i+cas];
    for (i=0; i<dn; i++) b[sn+i]=a[(2*i+1-cas)];
}

void dwt_deinterleave_v(int *a, int *b, int dn, int sn, int x, int cas) {
    int i;
    for (i=0; i<sn; i++) b[i*x]=a[2*i+cas];
    for (i=0; i<dn; i++) b[(sn+i)*x]=a[(2*i+1-cas)];
}
#else
void dwt_encode(char * out, int num) {
	int i, j, k;
	char *a = NULL;
	char *aj = NULL;
	char *bj = NULL;
	int w, l;

	w = WIDTH_TILE;
	l = NUM_RES - 1;
	a = out;
	
	for (i = 0; i < l; i++) {
		int rw;			// width of the resolution level computed 
		int rh;			// heigth of the resolution level computed                                 
		int rw1;		// width of the resolution level once lower than computed one
		int rh1;		// height of the resolution level once lower than computed one
		int cas_col;	// 0 = non inversion on horizontal filtering 1 = inversion between low-pass and high-pass filtering
		int cas_row;	// 0 = non inversion on vertical filtering 1 = inversion between low-pass and high-pass filtering
		int dn, sn;
		char bj_alloc[WIDTH_TILE>HEIGHT_TILE?WIDTH_TILE:HEIGHT_TILE];
			
		rw = RES_WIDTH[l - i + RES_OFFSET];
		rh = RES_HEIGTH[l - i + RES_OFFSET];
		rw1= RES_WIDTH[l - i - 1 + RES_OFFSET];
		rh1= RES_HEIGTH[l - i - 1 + RES_OFFSET];
		
		int p = num % HTILES;
		int q = num / HTILES;
		
		cas_row = (p * RES_WIDTH[l - i + RES_OFFSET]) % 2;
		cas_col = (q * RES_HEIGTH[l - i + RES_OFFSET]) % 2;
        
		sn = rh1;
		dn = rh - rh1;
		bj = bj_alloc;
		for (j = 0; j < rw; j++) {
			aj = a + j;
			for (k = 0; k < rh; k++)  bj[k] = aj[k*w];
			dwt_encode_1(bj, dn, sn, cas_col);
			dwt_deinterleave_v(bj, aj, dn, sn, w, cas_col);
		}
		
		sn = rw1;
		dn = rw - rw1;
		bj = bj_alloc;
		for (j = 0; j < rh; j++) {
			aj = a + j * w;
			for (k = 0; k < rw; k++)  bj[k] = aj[k];
			dwt_encode_1(bj, dn, sn, cas_row);
			dwt_deinterleave_h(bj, aj, dn, sn, cas_row);
		}
	}
}

void dwt_encode_1(char *a, int dn, int sn, int cas) {
	int i;
	int lim;
	
	if (!cas) {
		if ((dn > 0) || (sn > 1)) {	
			for (i = 0; i < dn; i++)
			{
				lim = D(i);
				lim -= (S_(i) + S_(i + 1)) >> 1;
				if (lim > 127) D(i) = 127;
				else if (lim < -128) D(i) = -128;
				else D(i) = lim;
			}
			for (i = 0; i < sn; i++)
			{
				lim = S(i);
				lim += (D_(i - 1) + D_(i) + 2) >> 2;
				if (lim > 127) S(i) = 127;
				else if (lim < -128) S(i) = -128;
				else S(i) = lim;
			}
		}
	} else {
		if (!sn && dn == 1)
			S(0) *= 2;
		else {
			for (i = 0; i < dn; i++)
			{
				lim = S(i);
				lim -= (DD_(i) + DD_(i - 1)) >> 1;
				if (lim > 127) S(i) = 127;
				else if (lim < -128) S(i) = -128;
				else S(i) = lim;
			}
			for (i = 0; i < sn; i++)
			{
				lim = D(i);
				lim += (SS_(i) + SS_(i + 1) + 2) >> 2;
				if (lim > 127) D(i) = 127;
				else if (lim < -128) D(i) = -128;
				else D(i) = lim;
			}
		}
	}
}

void dwt_deinterleave_h(char *a, char *b, int dn, int sn, int cas) {
	int i;
    for (i=0; i<sn; i++) b[i]=a[2*i+cas];
    for (i=0; i<dn; i++) b[sn+i]=a[(2*i+1-cas)];
}

void dwt_deinterleave_v(char *a, char *b, int dn, int sn, int x, int cas) {
    int i;
    for (i=0; i<sn; i++) b[i*x]=a[2*i+cas];
    for (i=0; i<dn; i++) b[(sn+i)*x]=a[(2*i+1-cas)];
}
#endif

