/*******************************************************************\

Copyright (c) 2008 Leiden University (LERC group at LIACS).
All rights reserved.

The use and distribution terms for this software are covered by the 
Common Public License 1.0 (http://opensource.org/licenses/cpl1.0.txt)
which can be found in the file LICENSE at the root of this distribution.
By using this software in any fashion, you are agreeing to be bound by 
the terms of this license.

You must not remove this notice, or any other, from this software.

\*******************************************************************/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "T2.h"

void mainT2(const code_pas_t * codepas_in, packet_t * packets_out)
{
	int len;
	unsigned char *dest;
	layers_t lay;
	res_t rest;
	infos_t info;
	infos_t * info_in;
	tree_comp_t tree;
	
	info_in = &info;
	info_in->compnum = codepas_in->c.numcomp;

	packets_out->num = codepas_in->c.num;
	packets_out->numcomp = codepas_in->c.numcomp;
	
	j2k_setup_encoder(info_in);
	tcd_init_encode(info_in,packets_out->num);

	tcd_rateallocate(info_in, &codepas_in->c,&lay,&tree, &codepas_in->p);

	//Rate-Distorsion Allocate
	dest = packets_out->data;
	len = (IMAGE_SIZE / NUMTILES) -2 ;

	packets_out->lenght = t2_encode_packets(packets_out->num, NUM_LAYERS, dest, len, CUR_TP_NUM,TP_POS, CUR_PINO,FINAL_PASS, info_in, &codepas_in->c, &lay,&rest,&tree, &codepas_in->p);
}

int tcd_rateallocate(infos_t * info, const codeblock_t * blocks_in, layers_t * lay, tree_comp_t * tree, const passes_t * passes_in) {
	int resno, bandno, precno, cblkno, passno, layno;
	double min, max;
	
	min = DBL_MAX;
	max = 0;

	lay->nbpix = 0;
	lay->nbpixc = 0;

	for (resno = 0; resno < NUM_RES; resno++) {

		int resnomax = resno ? NUM_BANDS : 1;
		for (bandno = 0; bandno < resnomax; bandno++) {
			for (precno = 0; precno < NUM_PRECS; precno++) {		
				tgt_create(&tree->incltree[resno][bandno][precno], info->cw[resno][bandno][precno],info->ch[resno][bandno][precno]);
				tgt_create(&tree->imsbtree[resno][bandno][precno], info->cw[resno][bandno][precno],info->ch[resno][bandno][precno]);

				for (cblkno = 0; cblkno < info->codebs[resno][bandno][precno]; cblkno++) {
											
					for (passno = 0; passno < passes_in->passnum[resno][bandno][precno][cblkno]; passno++) {

						int dr;
						double dd, rdslope;
						if (passno == 0) {
							dr = passes_in->passes[resno][bandno][precno][cblkno][passno].rate;
							dd = passes_in->passes[resno][bandno][precno][cblkno][passno].distortiondec;
						} else {
							dr = passes_in->passes[resno][bandno][precno][cblkno][passno].rate - passes_in->passes[resno][bandno][precno][cblkno][passno-1].rate;
							dd = passes_in->passes[resno][bandno][precno][cblkno][passno].distortiondec - passes_in->passes[resno][bandno][precno][cblkno][passno-1].distortiondec;
						}
						if (dr == 0) {
							continue;
						}
						rdslope = dd / dr;
						if (rdslope < min) {
							min = rdslope;
						}
						if (rdslope > max) {
							max = rdslope;
						}
					}

					lay->nbpix += (info->codebx1[resno][bandno][precno][cblkno] - info->codebx0[resno][bandno][precno][cblkno])
										* (info->codeby1[resno][bandno][precno][cblkno] - info->codeby0[resno][bandno][precno][cblkno]);
					lay->nbpixc += (info->codebx1[resno][bandno][precno][cblkno] - info->codebx0[resno][bandno][precno][cblkno])
										* (info->codeby1[resno][bandno][precno][cblkno] - info->codeby0[resno][bandno][precno][cblkno]);
				}
			}
		}
	}

	for (layno = 0; layno < NUM_LAYERS; layno++) {
		double goodthresh = 0;
		
		goodthresh = min;
		
		tcd_makelayer(layno, goodthresh, 1, info,lay,passes_in);
	}

	return true;
}

int t2_encode_packets(int tileno, int maxlayers, unsigned char *dest, int len, int tpnum, int tppos,int pino, J2K_T2_MODE t2_mode, infos_t * info, const codeblock_t * blocks_in, layers_t * lay, res_t * rest, tree_comp_t * tree, const passes_t * passes_in){
	unsigned char *c = dest;
	int e = 0;
	int num = 0;
	opj_pi_iterator_t pi[NUM_POCS + 1];
	
	pi_initialise_encode(pi, tileno, t2_mode, info,rest);

	pi_create_encode(pi, tileno,pino,tpnum,tppos,info); 
	
	while (pi_next(&pi[pino],rest)) {
		if (pi[pino].layno < maxlayers) {
			e = t2_encode_packet(&pi[pino], c, dest + len - c, tileno, info, blocks_in,lay,&num,tree,passes_in);
			if (e == -999) {
				break;
			} else {
				c += e;
			}
		}
	}

	if (e == -999) {
		return e;
	}

    return (c - dest);
}

void tcd_makelayer(int layno, double thresh, int final, const infos_t * info, layers_t * lay, const passes_t * passes_in) {
	int resno, bandno, precno, cblkno, passno;

	lay->distolayer[layno] = 0;

	for (resno = 0; resno < NUM_RES; resno++) {

		int resnomax = resno ? NUM_BANDS : 1;
		for (bandno = 0; bandno < resnomax; bandno++) {

			for (precno = 0; precno < NUM_PRECS; precno++) {

				for (cblkno = 0; cblkno < info->codebs[resno][bandno][precno]; cblkno++) {
					int n;
					if (layno == 0) {
						lay->numpasses[resno][bandno][precno][cblkno] = 0;
					}
					n = lay->numpasses[resno][bandno][precno][cblkno];
					for (passno = lay->numpasses[resno][bandno][precno][cblkno]; passno < passes_in->passnum[resno][bandno][precno][cblkno]; passno++) {
						int dr;
						double dd;

						if (n == 0) {
							dr = passes_in->passes[resno][bandno][precno][cblkno][passno].rate;
							dd = passes_in->passes[resno][bandno][precno][cblkno][passno].distortiondec;
						} else {
							dr = passes_in->passes[resno][bandno][precno][cblkno][passno].rate - passes_in->passes[resno][bandno][precno][cblkno][passno-1].rate;
							dd = passes_in->passes[resno][bandno][precno][cblkno][passno].distortiondec - passes_in->passes[resno][bandno][precno][cblkno][passno-1].distortiondec;
						}
						if (!dr) {
							if (dd != 0)
								n = passno + 1;
							continue;
						}
						if (dd / dr >= thresh)
							n = passno + 1;
					}
					lay->layer[resno][bandno][precno][cblkno][layno].numpasses = n - lay->numpasses[resno][bandno][precno][cblkno];
					
					if (!lay->layer[resno][bandno][precno][cblkno][layno].numpasses) {
						lay->layer[resno][bandno][precno][cblkno][layno].disto = 0;
						continue;
					}

					lay->layer[resno][bandno][precno][cblkno][layno].len = passes_in->passes[resno][bandno][precno][cblkno][n-1].rate;
					lay->layer[resno][bandno][precno][cblkno][layno].disto = passes_in->passes[resno][bandno][precno][cblkno][n-1].distortiondec;
					
					lay->distolayer[layno] += lay->layer[resno][bandno][precno][cblkno][layno].disto;
					
					if (final)
						lay->numpasses[resno][bandno][precno][cblkno] = n;
				}
			}
		}
	}
}

void pi_create_encode( opj_pi_iterator_t *pi, int tileno, int pino,int tpnum, int tppos,const infos_t * info){
	char *prog;
	int i,l;
	int incr_top=1,resetX=0;

	prog = (char *) j2k_convert_progression_order(PROG_ORDER);

	pi[pino].first = 1;
	pi[pino].poc.prg = PROG_ORDER;	
	
	if(!(TP_ON)){
		pi[pino].poc.resno0 = 0;
		pi[pino].poc.resno1 = NUM_RES;
		pi[pino].poc.compno0 = 0;
		pi[pino].poc.compno1 = NUM_COMPS;
		pi[pino].poc.layno0 = 0;
		pi[pino].poc.layno1 = NUM_LAYERS;
		pi[pino].poc.precno0 = 0;
		pi[pino].poc.precno1 = NUM_PRECS;
		pi[pino].poc.tx0 = info->tx0;
		pi[pino].poc.ty0 = info->ty0;
		pi[pino].poc.tx1 = info->tx1;
		pi[pino].poc.ty1 = info->ty1;
	}else {
		for(i=tppos+1;i<4;i++){
			switch(prog[i]){
			case 'R':
				pi[pino].poc.resno0 = 0;
				pi[pino].poc.resno1 = NUM_RES;
				break;
			case 'C':
				pi[pino].poc.compno0 = 0;
				pi[pino].poc.compno1 = NUM_COMPS;
				break;
			case 'L':
				pi[pino].poc.layno0 = 0;
				pi[pino].poc.layno1 = NUM_LAYERS;
				break;
			case 'P':
				switch(PROG_ORDER){
					case LRCP:
					case RLCP:
						pi[pino].poc.precno0 = 0;
						pi[pino].poc.precno1 = NUM_PRECS;
						break;
					default:
						pi[pino].poc.tx0 = info->tx0;
						pi[pino].poc.ty0 = info->ty0;
						pi[pino].poc.tx1 = info->tx1;
						pi[pino].poc.ty1 = info->ty1;
						break;
				}
				break;
			}
		}

		if(tpnum==0){
			for(i=tppos;i>=0;i--){
				switch(prog[i]){
						case 'C':
							pi->comp_t = 0;
							pi[pino].poc.compno0 = pi->comp_t;
							pi[pino].poc.compno1 = pi->comp_t+1;
							pi->comp_t+=1;
							break;
						case 'R':
							pi->res_t = 0;
							pi[pino].poc.resno0 = pi->res_t;
							pi[pino].poc.resno1 = pi->res_t+1;
							pi->res_t+=1;
							break;
						case 'L':
							pi->lay_t = 0;
							pi[pino].poc.layno0 = pi->lay_t;
							pi[pino].poc.layno1 = pi->lay_t+1;
							pi->lay_t+=1;
							break;
						case 'P':
							switch(PROG_ORDER){
								case LRCP:
								case RLCP:
									pi->prc_t = 0;
									pi[pino].poc.precno0 = pi->prc_t;
									pi[pino].poc.precno1 = pi->prc_t+1;
									pi->prc_t+=1; 
									break;
								default:
									pi->tx0_t = info->tx0;
									pi->ty0_t = info->ty0;
									pi[pino].poc.tx0 = pi->tx0_t;
									pi[pino].poc.tx1 = pi->tx0_t + SUBSAMPLING_DX - (pi->tx0_t % SUBSAMPLING_DX);
									pi[pino].poc.ty0 = pi->ty0_t;
									pi[pino].poc.ty1 = pi->ty0_t + SUBSAMPLING_DY - (pi->ty0_t % SUBSAMPLING_DY);
									pi->tx0_t = pi[pino].poc.tx1;
									pi->ty0_t = pi[pino].poc.ty1;
									break;
							}
							break;
				}
			}
			incr_top=1;
		}else{
			for(i=tppos;i>=0;i--){
				switch(prog[i]){
						case 'C':
							pi[pino].poc.compno0 = pi->comp_t-1;
							pi[pino].poc.compno1 = pi->comp_t;
							break;
						case 'R':
							pi[pino].poc.resno0 = pi->res_t-1;
							pi[pino].poc.resno1 = pi->res_t;
							break;
						case 'L':
							pi[pino].poc.layno0 = pi->lay_t-1;
							pi[pino].poc.layno1 = pi->lay_t;
							break;
						case 'P':
							switch(PROG_ORDER ){
								case LRCP:
								case RLCP:
									pi[pino].poc.precno0 = pi->prc_t-1;
									pi[pino].poc.precno1 = pi->prc_t;
									break;
								default:
									pi[pino].poc.tx0 = pi->tx0_t - SUBSAMPLING_DX - (pi->tx0_t % SUBSAMPLING_DX);
									pi[pino].poc.tx1 = pi->tx0_t ;
									pi[pino].poc.ty0 = pi->ty0_t - SUBSAMPLING_DY - (pi->ty0_t % SUBSAMPLING_DY);
									pi[pino].poc.ty1 = pi->ty0_t ;
									break;
							}
							break;
				}
				if(incr_top==1){
					switch(prog[i]){
							case 'R':
								if(pi->res_t== NUM_RES){
									l=pi_check_next_level(i-1,pi,tileno,pino,prog,info);
									if(l==1){
										pi->res_t = 0;
										pi[pino].poc.resno0 = pi->res_t;
										pi[pino].poc.resno1 = pi->res_t+1;
										pi->res_t+=1;
										incr_top=1;
									}else{
										incr_top=0;
									}
								}else{
									pi[pino].poc.resno0 = pi->res_t;
									pi[pino].poc.resno1 = pi->res_t+1;
									pi->res_t+=1;
									incr_top=0;
								}
								break;
							case 'C':
								if(pi->comp_t ==NUM_COMPS){
									l=pi_check_next_level(i-1,pi,tileno,pino,prog,info);
									if(l==1){
										pi->comp_t = 0;
										pi[pino].poc.compno0 = pi->comp_t;
										pi[pino].poc.compno1 = pi->comp_t+1;
										pi->comp_t+=1;
										incr_top=1;
									}else{
										incr_top=0;
									}
								}else{
									pi[pino].poc.compno0 = pi->comp_t;
									pi[pino].poc.compno1 = pi->comp_t+1;
									pi->comp_t+=1;
									incr_top=0;
								}
								break;	
							case 'L':
								if(pi->lay_t == NUM_LAYERS){
									l=pi_check_next_level(i-1,pi,tileno,pino,prog,info);
									if(l==1){
										pi->lay_t = 0;
										pi[pino].poc.layno0 = pi->lay_t;
										pi[pino].poc.layno1 = pi->lay_t+1;
										pi->lay_t+=1;
										incr_top=1;
									}else{
										incr_top=0;
									}
								}else{
									pi[pino].poc.layno0 = pi->lay_t;
									pi[pino].poc.layno1 = pi->lay_t+1;
									pi->lay_t+=1;
									incr_top=0;
								}
								break;
							case 'P':
								switch(PROG_ORDER){
									case LRCP:
									case RLCP:
										if(pi->prc_t == NUM_PRECS){
											l=pi_check_next_level(i-1,pi,tileno,pino,prog,info);
											if(l==1){
												pi->prc_t = 0;
												pi[pino].poc.precno0 = pi->prc_t;
												pi[pino].poc.precno1 = pi->prc_t+1;
												pi->prc_t+=1;
												incr_top=1;
											}else{
												incr_top=0;
											}
										}else{
											pi[pino].poc.precno0 = pi->prc_t;
											pi[pino].poc.precno1 = pi->prc_t+1;
											pi->prc_t+=1;
											incr_top=0;
										}
										break;
									default:
										if(pi->tx0_t >= info->tx1){
											if(pi->ty0_t >= info->ty1){
												l=pi_check_next_level(i-1,pi,tileno,pino,prog,info);
												if(l==1){
													pi->ty0_t = info->ty0;
													pi[pino].poc.ty0 = pi->ty0_t;
													pi[pino].poc.ty1 = pi->ty0_t + SUBSAMPLING_DY - (pi->ty0_t % SUBSAMPLING_DY);
													pi->ty0_t = pi[pino].poc.ty1;
													incr_top=1;resetX=1;
												}else{
													incr_top=0;resetX=0;
												}
											}else{
												pi[pino].poc.ty0 = pi->ty0_t;
												pi[pino].poc.ty1 = pi->ty0_t + SUBSAMPLING_DY - (pi->ty0_t % SUBSAMPLING_DY);
												pi->ty0_t = pi[pino].poc.ty1;
												incr_top=0;resetX=1;
											}
											if(resetX==1){
												pi->tx0_t = info->tx0;
												pi[pino].poc.tx0 = pi->tx0_t;
												pi[pino].poc.tx1 = pi->tx0_t + SUBSAMPLING_DX- (pi->tx0_t % SUBSAMPLING_DX);
												pi->tx0_t = pi[pino].poc.tx1;
											}
										}else{
											pi[pino].poc.tx0 = pi->tx0_t;
											pi[pino].poc.tx1 = pi->tx0_t + SUBSAMPLING_DX- (pi->tx0_t % SUBSAMPLING_DX);
											pi->tx0_t = pi[pino].poc.tx1;
											incr_top=0;
										}
										break;
								}
								break;
					}
				}
			}
		}
	}
}

int pi_check_next_level(int pos, opj_pi_iterator_t *pi,int tileno, int pino, char *prog, const infos_t * info){
	int i,l;

	if(pos>=0){
		for(i=pos;pos>=0;i--){
			switch(prog[i]){
		case 'R':
			if(pi->res_t== NUM_RES){
				l=pi_check_next_level(pos-1,pi,tileno,pino,prog,info);
				if(l==1){
					return 1;
				}else{
					return 0;
				}
			}else{
				return 1;
			}
			break;
		case 'C':
			if(pi->comp_t==NUM_COMPS){
				l=pi_check_next_level(pos-1,pi,tileno,pino,prog,info);
				if(l==1){
					return 1;
				}else{
					return 0;
				}
			}else{
				return 1;
			}
			break;
		case 'L':
			if(pi->lay_t== NUM_LAYERS){
				l=pi_check_next_level(pos-1,pi,tileno,pino,prog,info);
				if(l==1){
					return 1;
				}else{
					return 0;
				}
			}else{
				return 1;
			}
			break;
		case 'P':
			switch(PROG_ORDER){
				case LRCP||RLCP:
					if(pi->prc_t == NUM_PRECS){
						l=pi_check_next_level(i-1,pi,tileno,pino,prog,info);
						if(l==1){
							return 1;
						}else{
							return 0;
						}
					}else{
						return 1;
					}
					break;
			default:
				if(pi->tx0_t == info->tx1){
					//TY 
					if(pi->ty0_t == info->ty1){
						l=pi_check_next_level(i-1,pi,tileno,pino,prog,info);
						if(l==1){
							return 1;
						}else{
							return 0;
						}
					}else{
						return 1;
					}//TY
				}else{
					return 1;
				}
				break;
			}//end case P
		}//end switch
		}//end for
	}//end if
	return 0;
}

int pi_next(opj_pi_iterator_t * pi,res_t * rest) {
	switch (PROG_ORDER) {
		case LRCP:
			return pi_next_lrcp(pi);
		case RLCP:
			return pi_next_rlcp(pi);
		case RPCL:
			return pi_next_rpcl(pi,rest);
		case PCRL:
			return pi_next_pcrl(pi,rest);
		case CPRL:
			return pi_next_cprl(pi,rest);
		case PROG_UNKNOWN:
			return false;
	}
	
	return false;
}

int pi_next_lrcp(opj_pi_iterator_t * pi) {
	opj_pi_comp_t *comp = NULL;
	opj_pi_resolution_t *res = NULL;
	long index = 0;
	
	if (!pi->first) {
		comp = &pi->comps;
		res = &comp->resolutions[pi->resno];
		goto LABEL_SKIP;
	} else {
		pi->first = 0;
	}

	for (pi->layno = pi->poc.layno0; pi->layno < pi->poc.layno1; pi->layno++) {
		for (pi->resno = pi->poc.resno0; pi->resno < pi->poc.resno1;
		pi->resno++) {
			comp = &pi->comps;
			if (pi->resno >= comp->numresolutions) {
				continue;
			}
			res = &comp->resolutions[pi->resno];
			if (!pi->tp_on){
				pi->poc.precno1 = NUM_PRECS;
			}
			for (pi->precno = pi->poc.precno0; pi->precno < pi->poc.precno1; pi->precno++) {
				index = pi->layno * pi->step_l + pi->resno * pi->step_r + pi->compno * pi->step_c + pi->precno * pi->step_p;
				if (!pi->include[index]) {
					pi->include[index] = 1;
					return true;
				}
LABEL_SKIP:;
			}
		}
	}
	
	return false;
}

int pi_next_rlcp(opj_pi_iterator_t * pi) {
	opj_pi_comp_t *comp = NULL;
	opj_pi_resolution_t *res = NULL;
	long index = 0;

	if (!pi->first) {
		comp = &pi->comps;
		res = &comp->resolutions[pi->resno];
		goto LABEL_SKIP;
	} else {
		pi->first = 0;
	}

	for (pi->resno = pi->poc.resno0; pi->resno < pi->poc.resno1; pi->resno++) {
		for (pi->layno = pi->poc.layno0; pi->layno < pi->poc.layno1; pi->layno++) {
			//for (pi->compno = pi->poc.compno0; pi->compno < pi->poc.compno1; pi->compno++) {
				comp = &pi->comps;
				if (pi->resno >= comp->numresolutions) {
					continue;
				}
				res = &comp->resolutions[pi->resno];
				if(!pi->tp_on){
					pi->poc.precno1 = NUM_PRECS;
				}
				for (pi->precno = pi->poc.precno0; pi->precno < pi->poc.precno1; pi->precno++) {
					index = pi->layno * pi->step_l + pi->resno * pi->step_r + pi->compno * pi->step_c + pi->precno * pi->step_p;
					if (!pi->include[index]) {
						pi->include[index] = 1;
						return true;
					}
LABEL_SKIP:;
				}
		}
	}
	
	return false;
}

int pi_next_rpcl(opj_pi_iterator_t * pi, res_t * rest) {
	opj_pi_comp_t *comp = NULL;
	opj_pi_resolution_t *res = NULL;
	long index = 0;

	if (!pi->first) {
		goto LABEL_SKIP;
	} else {
		int resno;
		pi->first = 0;
		pi->dx = 0;
		pi->dy = 0;
		
		comp = &pi->comps;
		for (resno = 0; resno < comp->numresolutions; resno++) {
			int dx, dy;
			res = &comp->resolutions[resno];
			dx = comp->dx * (1 << (res->pdx + comp->numresolutions - 1 - resno));
			dy = comp->dy * (1 << (res->pdy + comp->numresolutions - 1 - resno));
			pi->dx = !pi->dx ? dx : int_min(pi->dx, dx);
			pi->dy = !pi->dy ? dy : int_min(pi->dy, dy);
		}
	}
if (!pi->tp_on){
			pi->poc.ty0 = pi->ty0;
			pi->poc.tx0 = pi->tx0;
			pi->poc.ty1 = pi->ty1;
			pi->poc.tx1 = pi->tx1;
		}
	for (pi->resno = pi->poc.resno0; pi->resno < pi->poc.resno1; pi->resno++) {
		for (pi->y = pi->poc.ty0; pi->y < pi->poc.ty1; pi->y += pi->dy - (pi->y % pi->dy)) {
			for (pi->x = pi->poc.tx0; pi->x < pi->poc.tx1; pi->x += pi->dx - (pi->x % pi->dx)) {
				for (pi->compno = pi->poc.compno0; pi->compno < pi->poc.compno1; pi->compno++) {
					int levelno;
					int trx0, try0;
					int trx1, try1;
					int rpx, rpy;
					int prci, prcj;
					comp = &pi->comps;
					if (pi->resno >= comp->numresolutions) {
						continue;
					}
					res = &comp->resolutions[pi->resno];
					levelno = comp->numresolutions - 1 - pi->resno;
					trx0 = int_ceildiv(pi->tx0, comp->dx << levelno);
					try0 = int_ceildiv(pi->ty0, comp->dy << levelno);
					trx1 = int_ceildiv(pi->tx1, comp->dx << levelno);
					try1 = int_ceildiv(pi->ty1, comp->dy << levelno);
					rpx = rest->pdx[pi->resno] + levelno;
					rpy = rest->pdy[pi->resno] + levelno;
					if (!((pi->y % (comp->dy << rpy) == 0) || ((pi->y == pi->ty0) && ((try0 << levelno) % (1 << rpx))))){
						continue;	
					}
					if (!((pi->x % (comp->dx << rpx) == 0) || ((pi->x == pi->tx0) && ((trx0 << levelno) % (1 << rpx))))){
						continue; 
					}
					
					if ((rest->pw[pi->resno]==0)||(rest->pw[pi->resno]==0)) continue;
					
					if ((trx0==trx1)||(try0==try1)) continue;
					
					prci = int_floordivpow2(int_ceildiv(pi->x, comp->dx << levelno), rest->pdx[pi->resno]) 
						 - int_floordivpow2(trx0, rest->pdx[pi->resno]);
					prcj = int_floordivpow2(int_ceildiv(pi->y, comp->dy << levelno), rest->pdy[pi->resno]) 
						 - int_floordivpow2(try0, rest->pdy[pi->resno]);
					pi->precno = prci + prcj * rest->pw[pi->resno];
					for (pi->layno = pi->poc.layno0; pi->layno < pi->poc.layno1; pi->layno++) {
						index = pi->layno * pi->step_l + pi->resno * pi->step_r + pi->compno * pi->step_c + pi->precno * pi->step_p;
						if (!pi->include[index]) {
							pi->include[index] = 1;
							return true;
						}
LABEL_SKIP:;
					}
			}
		}
	}
	}
	
	return false;
}

int pi_next_pcrl(opj_pi_iterator_t * pi, res_t * rest) {
	opj_pi_comp_t *comp = NULL;
	opj_pi_resolution_t *res = NULL;
	long index = 0;

	if (!pi->first) {
		comp = &pi->comps;
		goto LABEL_SKIP;
	} else {
		int resno;
		pi->first = 0;
		pi->dx = 0;
		pi->dy = 0;
		comp = &pi->comps;
		for (resno = 0; resno < comp->numresolutions; resno++) {
			int dx, dy;
			res = &comp->resolutions[resno];
			dx = comp->dx * (1 << (res->pdx + comp->numresolutions - 1 - resno));
			dy = comp->dy * (1 << (res->pdy + comp->numresolutions - 1 - resno));
			pi->dx = !pi->dx ? dx : int_min(pi->dx, dx);
			pi->dy = !pi->dy ? dy : int_min(pi->dy, dy);
		}
	}
	if (!pi->tp_on){
			pi->poc.ty0 = pi->ty0;
			pi->poc.tx0 = pi->tx0;
			pi->poc.ty1 = pi->ty1;
			pi->poc.tx1 = pi->tx1;
		}
	for (pi->y = pi->poc.ty0; pi->y < pi->poc.ty1; pi->y += pi->dy - (pi->y % pi->dy)) {
		for (pi->x = pi->poc.tx0; pi->x < pi->poc.tx1; pi->x += pi->dx - (pi->x % pi->dx)) {
			//for (pi->compno = pi->poc.compno0; pi->compno < pi->poc.compno1; pi->compno++) {
				comp = &pi->comps;
				for (pi->resno = pi->poc.resno0; pi->resno < int_min(pi->poc.resno1, comp->numresolutions); pi->resno++) {
					int levelno;
					int trx0, try0;
					int trx1, try1;
					int rpx, rpy;
					int prci, prcj;
					res = &comp->resolutions[pi->resno];
					levelno = comp->numresolutions - 1 - pi->resno;
					trx0 = int_ceildiv(pi->tx0, comp->dx << levelno);
					try0 = int_ceildiv(pi->ty0, comp->dy << levelno);
					trx1 = int_ceildiv(pi->tx1, comp->dx << levelno);
					try1 = int_ceildiv(pi->ty1, comp->dy << levelno);
					rpx = res->pdx + levelno;
					rpy = res->pdy + levelno;
					if (!((pi->y % (comp->dy << rpy) == 0) || ((pi->y == pi->ty0) && ((try0 << levelno) % (1 << rpx))))){
						continue;	
					}
					if (!((pi->x % (comp->dx << rpx) == 0) || ((pi->x == pi->tx0) && ((trx0 << levelno) % (1 << rpx))))){
						continue; 
					}
					
					if ((rest->pw[pi->resno]==0)||(rest->pw[pi->resno]==0)) continue;
					
					if ((trx0==trx1)||(try0==try1)) continue;
					
					prci = int_floordivpow2(int_ceildiv(pi->x, comp->dx << levelno), rest->pdx[pi->resno]) 
						 - int_floordivpow2(trx0, rest->pdx[pi->resno]);
					prcj = int_floordivpow2(int_ceildiv(pi->y, comp->dy << levelno), rest->pdy[pi->resno]) 
						 - int_floordivpow2(try0, rest->pdx[pi->resno]);
					pi->precno = prci + prcj * rest->pw[pi->resno];
					for (pi->layno = pi->poc.layno0; pi->layno < pi->poc.layno1; pi->layno++) {
						index = pi->layno * pi->step_l + pi->resno * pi->step_r + pi->compno * pi->step_c + pi->precno * pi->step_p;
						if (!pi->include[index]) {
							pi->include[index] = 1;
							return true;
						}	
LABEL_SKIP:;
					}
				}
			}
		}
	
	return false;
}

int pi_next_cprl(opj_pi_iterator_t * pi,res_t * rest) {
	opj_pi_comp_t *comp = NULL;
	opj_pi_resolution_t *res = NULL;
	long index = 0;

	if (!pi->first) {
		comp = &pi->comps;
		goto LABEL_SKIP;
	} else {
		pi->first = 0;
	}

	int resno;
	comp = &pi->comps;
	pi->dx = 0;
	pi->dy = 0;
	for (resno = 0; resno < comp->numresolutions; resno++) {
		int dx, dy;
		res = &comp->resolutions[resno];
		dx = comp->dx * (1 << (res->pdx + comp->numresolutions - 1 - resno));
		dy = comp->dy * (1 << (res->pdy + comp->numresolutions - 1 - resno));
		pi->dx = !pi->dx ? dx : int_min(pi->dx, dx);
		pi->dy = !pi->dy ? dy : int_min(pi->dy, dy);
	}
	if (!pi->tp_on){
		pi->poc.ty0 = pi->ty0;
		pi->poc.tx0 = pi->tx0;
		pi->poc.ty1 = pi->ty1;
		pi->poc.tx1 = pi->tx1;
	}
	for (pi->y = pi->poc.ty0; pi->y < pi->poc.ty1; pi->y += pi->dy - (pi->y % pi->dy)) {
		for (pi->x = pi->poc.tx0; pi->x < pi->poc.tx1; pi->x += pi->dx - (pi->x % pi->dx)) {
			for (pi->resno = pi->poc.resno0; pi->resno < int_min(pi->poc.resno1, comp->numresolutions); pi->resno++) {
				int levelno;
				int trx0, try0;
				int trx1, try1;
				int rpx, rpy;
				int prci, prcj;
				res = &comp->resolutions[pi->resno];
				levelno = comp->numresolutions - 1 - pi->resno;
				trx0 = int_ceildiv(pi->tx0, comp->dx << levelno);
				try0 = int_ceildiv(pi->ty0, comp->dy << levelno);
				trx1 = int_ceildiv(pi->tx1, comp->dx << levelno);
				try1 = int_ceildiv(pi->ty1, comp->dy << levelno);
				rpx = res->pdx + levelno;
				rpy = res->pdy + levelno;
				if (!((pi->y % (comp->dy << rpy) == 0) || ((pi->y == pi->ty0) && ((try0 << levelno) % (1 << rpx))))){
					continue;	
				}
				if (!((pi->x % (comp->dx << rpx) == 0) || ((pi->x == pi->tx0) && ((trx0 << levelno) % (1 << rpx))))){
					continue; 
				}
				
				if ((rest->pw[pi->resno]==0)||(rest->pw[pi->resno]==0)) continue;
				
				if ((trx0==trx1)||(try0==try1)) continue;
				
				prci = int_floordivpow2(int_ceildiv(pi->x, comp->dx << levelno), rest->pdx[pi->resno]) 
					 - int_floordivpow2(trx0, rest->pdx[pi->resno]);
				prcj = int_floordivpow2(int_ceildiv(pi->y, comp->dy << levelno), rest->pdy[pi->resno]) 
					 - int_floordivpow2(try0, rest->pdy[pi->resno]);
				pi->precno = prci + prcj * rest->pw[pi->resno];
				for (pi->layno = pi->poc.layno0; pi->layno < pi->poc.layno1; pi->layno++) {
					index = pi->layno * pi->step_l + pi->resno * pi->step_r + pi->compno * pi->step_c + pi->precno * pi->step_p;
					if (!pi->include[index]) {
						pi->include[index] = 1;
						return true;
					}
LABEL_SKIP:;
				}
			}
		}
	}
	
	return false;
}

int t2_encode_packet(opj_pi_iterator_t *pi, unsigned char *dest, int length, int tileno, infos_t * info, const codeblock_t * blocks_in, layers_t * lay,int * num, tree_comp_t * tree, const passes_t * passes_in) {
	int bandno, cblkno;
	unsigned char sop[6], eph[2];
	unsigned char *c = dest;

	//int compno = pi->compno;	// component value
	int resno  = pi->resno;		// resolution level value
	int precno = pi->precno;	// precinct value
	int layno  = pi->layno;		// quality layer value 

	opj_bio_t *bio = NULL;

	int numpasses[NUM_BANDS][NUM_PRECS][NUM_CODEBS];

	precno = pi->precno;
	
	if (CSTY & J2K_CP_CSTY_SOP) {
		sop[0] = 255;
		sop[1] = 145;
		sop[2] = 0;
		sop[3] = 4;
		sop[4] = (*num % 65536) / 256;
		sop[5] = (*num % 65536) % 256;
		memcpy(c, sop, 6);
		c += 6;
	}
	
	if (!layno) {
		int resnomax = resno ? NUM_BANDS : 1;
		for (bandno = 0; bandno < resnomax; bandno++) {

			tgt_reset(&tree->incltree[resno][bandno][precno]);
			tgt_reset(&tree->imsbtree[resno][bandno][precno]);
			for (cblkno = 0; cblkno < info->codebs[resno][bandno][precno]; cblkno++) {
				numpasses[bandno][precno][cblkno] = 0;
				tgt_setvalue(&tree->imsbtree[resno][bandno][precno], cblkno, info->numbpsb[resno][bandno] - passes_in->numbps[resno][bandno][precno][cblkno]);
			}
		}
	}
	
	opj_bio_t bio_alloc;
	bio = &bio_alloc;
	bio_init_enc(bio, c, length);
	bio_write(bio, 1, 1);
	
	// Writing Packet header
	int resnomax = resno ? NUM_BANDS : 1;
	for (bandno = 0; bandno < resnomax; bandno++) {

		for (cblkno = 0; cblkno < info->codebs[resno][bandno][precno]; cblkno++) {

			if (!numpasses[bandno][precno][cblkno] && lay->layer[resno][bandno][precno][cblkno][layno].numpasses) {
				tgt_setvalue(&tree->incltree[resno][bandno][precno], cblkno, layno);
			}
		}
		for (cblkno = 0; cblkno < info->codebs[resno][bandno][precno]; cblkno++) {

			int increment = 0;
			int nump = 0;
			int len = 0, passno;
			int numlenbits = 0;
			
			// cblk inclusion bits
			if (!numpasses[bandno][precno][cblkno]) {
				tgt_encode(bio, &tree->incltree[resno][bandno][precno], cblkno, layno + 1);
			} else {
				bio_write(bio, lay->numpasses[resno][bandno][precno][cblkno] != 0, 1);
			}
			// if cblk not included, go to the next cblk
			if (!lay->layer[resno][bandno][precno][cblkno][layno].numpasses) {
				continue;
			}
			// if first instance of cblk --> zero bit-planes information
			if (!numpasses[bandno][precno][cblkno]) {
				numlenbits = 3;
				tgt_encode(bio,&tree->imsbtree[resno][bandno][precno], cblkno, 999);
			}
			// number of coding passes included
			t2_putnumpasses(bio, lay->layer[resno][bandno][precno][cblkno][layno].numpasses);
			
			// computation of the increase of the length indicator and insertion in the header
			for (passno = numpasses[bandno][precno][cblkno]; passno < numpasses[bandno][precno][cblkno] + lay->layer[resno][bandno][precno][cblkno][layno].numpasses; passno++) {
				nump++;
				len += passes_in->passes[resno][bandno][precno][cblkno][passno].len;
				if (passes_in->passes[resno][bandno][precno][cblkno][passno].term || passno == (numpasses[bandno][precno][cblkno] + lay->layer[resno][bandno][precno][cblkno][layno].numpasses) - 1) {
					increment = int_max(increment, int_floorlog2(len) + 1 - (numlenbits + int_floorlog2(nump)));
					len = 0;
					nump = 0;
				}
			}
			t2_putcommacode(bio, increment);

			// computation of the new Length indicator
			numlenbits += increment;

			// insertion of the codeword segment length
			for (passno = numpasses[bandno][precno][cblkno]; passno < numpasses[bandno][precno][cblkno] + lay->layer[resno][bandno][precno][cblkno][layno].numpasses; passno++) {
				nump++;
				len += passes_in->passes[resno][bandno][precno][cblkno][passno].len;
				if (passes_in->passes[resno][bandno][precno][cblkno][passno].term || passno == (numpasses[bandno][precno][cblkno] + lay->layer[resno][bandno][precno][cblkno][layno].numpasses) - 1) {
					bio_write(bio, len, numlenbits + int_floorlog2(nump));
					len = 0;
					nump = 0;
				}
			}
		}
	}
	
	if (bio_flush(bio)) {
		return -999;		// modified to eliminate longjmp !!
	}
	
	c += bio_numbytes(bio);

	

	if (CSTY & J2K_CP_CSTY_EPH) {
		eph[0] = 255;
		eph[1] = 146;
		memcpy(c, eph, 2);
		c += 2;
	}
	
	// Writing the packet body
	resnomax = resno ? NUM_BANDS : 1;
	for (bandno = 0; bandno < resnomax; bandno++) {

		for (cblkno = 0; cblkno < info->codebs[resno][bandno][precno]; cblkno++) {

			if (!lay->layer[resno][bandno][precno][cblkno][layno].numpasses) {
				continue;
			}
			if (c + lay->layer[resno][bandno][precno][cblkno][layno].len > dest + length) {
				return -999;
			}
			
			int k;
			for (k = 0; k < lay->layer[resno][bandno][precno][cblkno][layno].len; k ++)
			{				
				*(c++) = (int) blocks_in->data[resno][bandno][precno][cblkno][k];	
			}
			numpasses[bandno][precno][cblkno] += lay->layer[resno][bandno][precno][cblkno][layno].numpasses;
			
		}
	}
	return (c - dest);
}

void tgt_setvalue(opj_tgt_tree_t *tree, int leafno, int value) {
	opj_tgt_node_t *node;
	node = &tree->nodes[leafno];
	while (node && node->value > value) {
		node->value = value;
		node = node->parent;
	}
}

void bio_init_enc(opj_bio_t *bio, unsigned char *bp, int len) {
	bio->start = bp;
	bio->end = bp + len;
	bio->bp = bp;
	bio->buf = 0;
	bio->ct = 8;
}

void bio_write(opj_bio_t *bio, int v, int n) {
	int i;
	for (i = n - 1; i >= 0; i--) {
		bio_putbit(bio, (v >> i) & 1);
	}
}

void bio_putbit(opj_bio_t *bio, int b) {
	if (bio->ct == 0) {
		bio_byteout(bio);
	}
	bio->ct--;
	bio->buf |= b << bio->ct;
}

void tgt_encode(opj_bio_t *bio, opj_tgt_tree_t *tree, int leafno, int threshold) {
	opj_tgt_node_t *stk[31];
	opj_tgt_node_t **stkptr;
	opj_tgt_node_t *node;
	int low;

	stkptr = stk;
	node = &tree->nodes[leafno];
	while (node->parent) {
		*stkptr++ = node;
		node = node->parent;
	}
	
	low = 0;
	for (;;) {
		if (low > node->low) {
			node->low = low;
		} else {
			low = node->low;
		}
		
		while (low < threshold) {
			if (low >= node->value) {
				if (!node->known) {
					bio_write(bio, 1, 1);
					node->known = 1;
				}
				break;
			}
			bio_write(bio, 0, 1);
			++low;
		}
		
		node->low = low;
		if (stkptr == stk)
			break;
		node = *--stkptr;
	}
}

void t2_putnumpasses(opj_bio_t *bio, int n) {
	if (n == 1) {
		bio_write(bio, 0, 1);
	} else if (n == 2) {
		bio_write(bio, 2, 2);
	} else if (n <= 5) {
		bio_write(bio, 0xc | (n - 3), 4);
	} else if (n <= 36) {
		bio_write(bio, 0x1e0 | (n - 6), 9);
	} else if (n <= 164) {
		bio_write(bio, 0xff80 | (n - 37), 16);
	}
}

void t2_putcommacode(opj_bio_t *bio, int n) {
	while (--n >= 0) {
		bio_write(bio, 1, 1);
	}
	bio_write(bio, 0, 1);
}

int bio_flush(opj_bio_t *bio) {
	bio->ct = 0;
	if (bio_byteout(bio)) {
		return 1;
	}
	if (bio->ct == 7) {
		bio->ct = 0;
		if (bio_byteout(bio)) {
			return 1;
		}
	}
	return 0;
}

int bio_byteout(opj_bio_t *bio) {
	bio->buf = (bio->buf << 8) & 0xffff;
	bio->ct = bio->buf == 0xff00 ? 7 : 8;
	if (bio->bp >= bio->end) {
		return 1;
	}
	*bio->bp++ = bio->buf >> 8;
	return 0;
}

int bio_numbytes(opj_bio_t *bio) {
	return (bio->bp - bio->start);
}

void pi_initialise_encode(opj_pi_iterator_t *pi, int tileno, J2K_T2_MODE t2_mode, const infos_t * info, res_t * rest){ 
	int pino;
	int resno;
	int maxres = 0;
	int maxprec = 0;
	size_t array_size;
	
	pi->tp_on = TP_ON;
	pi->compno = info->compnum;
	
	for(pino = 0;pino < NUM_POCS + 1 ; pino ++){

		pi[pino].tx0 = info->tx0;
		pi[pino].ty0 = info->ty0;
		pi[pino].tx1 = info->tx1;
		pi[pino].ty1 = info->ty1;
		pi[pino].numcomps = NUM_COMPS;
		
		array_size = sizeof(opj_pi_comp_t);
		memset(&pi[pino].comps, 0, array_size);
		
		int tcx0, tcy0, tcx1, tcy1;
		opj_pi_comp_t *comp = &pi[pino].comps;
		comp->dx = SUBSAMPLING_DX;
		comp->dy = SUBSAMPLING_DY;
		comp->numresolutions = NUM_RES;

		tcx0 = int_ceildiv(pi[pino].tx0, SUBSAMPLING_DX);
		tcy0 = int_ceildiv(pi[pino].ty0, SUBSAMPLING_DY);
		tcx1 = int_ceildiv(pi[pino].tx1, SUBSAMPLING_DX);
		tcy1 = int_ceildiv(pi[pino].ty1, SUBSAMPLING_DY);
		
		if (NUM_RES > maxres) {
			maxres = NUM_RES;
		}
		
		for (resno = 0; resno < NUM_RES; resno++) {
			int levelno;
			int rx0, ry0, rx1, ry1;
			int px0, py0, px1, py1;
			
			rest->pdx[resno] = 15;
			rest->pdy[resno] = 15;
			
			levelno = NUM_COMPS - 1 - resno;
			rx0 = int_ceildivpow2(tcx0, levelno);
			ry0 = int_ceildivpow2(tcy0, levelno);
			rx1 = int_ceildivpow2(tcx1, levelno);
			ry1 = int_ceildivpow2(tcy1, levelno);
			px0 = int_floordivpow2(rx0, rest->pdx[resno]) << rest->pdx[resno];
			py0 = int_floordivpow2(ry0, rest->pdy[resno]) << rest->pdy[resno];
			px1 = int_ceildivpow2(rx1, rest->pdx[resno]) << rest->pdx[resno];
			py1 = int_ceildivpow2(ry1, rest->pdy[resno]) << rest->pdy[resno];
			rest->pw[resno] = (rx0==rx1)?0:((px1 - px0) >> rest->pdx[resno]);
			rest->ph[resno] = (ry0==ry1)?0:((py1 - py0) >> rest->pdy[resno]);

			if (rest->pw[resno] * rest->ph[resno] > maxprec) {
				maxprec = rest->pw[resno] * rest->ph[resno];
			}
		}
		
		pi[pino].step_p = 1;
		pi[pino].step_c = maxprec * pi[pino].step_p;
		pi[pino].step_r = pi[pino].step_c;
		pi[pino].step_l = maxres * pi[pino].step_r;
		
		//opj_pi_comp_t *comp = &pi->comps[compno];
		for (resno = 0; resno < NUM_RES; resno++) {
			int dx, dy;
			opj_pi_resolution_t *res = &comp->resolutions[resno];
			dx = comp->dx * (1 << (res->pdx + comp->numresolutions - 1 - resno));
			dy = comp->dy * (1 << (res->pdy + comp->numresolutions - 1 - resno));
			pi[pino].dx = !pi->dx ? dx : int_min(pi->dx, dx);
			pi[pino].dy = !pi->dy ? dy : int_min(pi->dy, dy);
		}

		if (pino == 0) {
			int k;
			short int *init = pi[pino].include;
			for (k = 0; k < INCLUDESIZE; k++)
			{
				*(init++) = 0;
			}
		}
		else {
			int k;
			for (k = 0; k < INCLUDESIZE; k++)
			{
				*(pi[pino].include + k) = *(pi[pino - 1].include + k);
			}
		}
	}
}

int int_max(int a, int b) {
	return (a > b) ? a : b;
}

void tgt_create(opj_tgt_tree_t *tree, int numleafsh, int numleafsv) {
	int nplh[32];
	int nplv[32];
	opj_tgt_node_t *node;
	opj_tgt_node_t *parentnode;
	opj_tgt_node_t *parentnode0;
	int i, j, k;
	int numlvls;
	int n;

	tree->numleafsh = numleafsh;
	tree->numleafsv = numleafsv;

	numlvls = 0;
	nplh[0] = numleafsh;
	nplv[0] = numleafsv;
	tree->numnodes = 0;

	do {
		n = nplh[numlvls] * nplv[numlvls];
		nplh[numlvls + 1] = (nplh[numlvls] + 1) / 2;
		nplv[numlvls + 1] = (nplv[numlvls] + 1) / 2;
		tree->numnodes += n;
		++numlvls;
	} while (n > 1);

	node = tree->nodes;
	parentnode = &tree->nodes[tree->numleafsh * tree->numleafsv];
	parentnode0 = parentnode;
	
	for (i = 0; i < numlvls - 1; ++i) {
		for (j = 0; j < nplv[i]; ++j) {
			k = nplh[i];
			while (--k >= 0) {
				node->parent = parentnode;
				++node;
				if (--k >= 0) {
					node->parent = parentnode;
					++node;
				}
				++parentnode;
			}
			if ((j & 1) || j == nplv[i] - 1) {
				parentnode0 = parentnode;
			} else {
				parentnode = parentnode0;
				parentnode0 += nplh[i];
			}
		}
	}
	node->parent = 0;
	
	tgt_reset(tree);
}

void tgt_reset(opj_tgt_tree_t *tree) {
	int i;

	if (NULL == tree)
		return;
	
	for (i = 0; i < tree->numnodes; i++) {
		tree->nodes[i].value = 999;
		tree->nodes[i].low = 0;
		tree->nodes[i].known = 0;
	}
}

j2k_prog_order_t j2k_prog_order_list[] = {
	{CPRL, "CPRL"},
	{LRCP, "LRCP"},
	{PCRL, "PCRL"},
	{RLCP, "RLCP"},
	{RPCL, "RPCL"},
	{PROG_UNKNOWN, ""}
};


char *j2k_convert_progression_order(OPJ_PROG_ORDER prg_order){
	j2k_prog_order_t *po;
	for(po = j2k_prog_order_list; po->enum_prog != -1; po++ ){
		if(po->enum_prog == prg_order){
			break;
		}
	}
	return po->str_prog;
}

void tcd_init_encode(infos_t * info_out,int num) {
	int  resno, bandno, precno, cblkno;
	int p = num % HTILES;
	int q = num / HTILES;
	
	info_out->tx0 = int_max(0 + p * WIDTH_TILE, 0);
	info_out->ty0 = int_max(0 + q * HEIGHT_TILE, 0);
	
	info_out->tx1 = int_min(0 + (p + 1) * WIDTH_TILE, WIDTH_IMAGE);
	info_out->ty1 = int_min(0 + (q + 1) * HEIGHT_TILE, HEIGHT_IMAGE);
			
	// border of each tile component (global)
	int tx0 = int_ceildiv(info_out->tx0, SUBSAMPLING_DX);
	int ty0 = int_ceildiv(info_out->ty0, SUBSAMPLING_DY);
	int tx1 = int_ceildiv(info_out->tx1, SUBSAMPLING_DX);
	int ty1 = int_ceildiv(info_out->ty1, SUBSAMPLING_DY);
							
	for (resno = 0; resno < NUM_RES; resno++) {
		int pdx, pdy;
		int levelno = NUM_RES - 1 - resno;
		int tlprcxstart, tlprcystart, brprcxend, brprcyend;
		int tlcbgxstart, tlcbgystart, brcbgxend, brcbgyend;
		int cbgwidthexpn, cbgheightexpn;
		int cblkwidthexpn, cblkheightexpn;
			
		// border for each resolution level (global)
		int rx0 = int_ceildivpow2(tx0, levelno);
		int ry0 = int_ceildivpow2(ty0, levelno);
		int rx1 = int_ceildivpow2(tx1, levelno);
		int ry1 = int_ceildivpow2(ty1, levelno);	

		pdx = 15;
		pdy = 15;

		tlprcxstart = int_floordivpow2(rx0, pdx) << pdx;
		tlprcystart = int_floordivpow2(ry0, pdy) << pdy;
		brprcxend = int_ceildivpow2(rx1, pdx) << pdx;
		brprcyend = int_ceildivpow2(ry1, pdy) << pdy;
			
		int rpw = (brprcxend - tlprcxstart) >> pdx;
			
		if (resno == 0) {
			tlcbgxstart = tlprcxstart;
			tlcbgystart = tlprcystart;
			brcbgxend = brprcxend;
			brcbgyend = brprcyend;
			cbgwidthexpn = pdx;
			cbgheightexpn = pdy;
		} else {
			tlcbgxstart = int_ceildivpow2(tlprcxstart, 1);
			tlcbgystart = int_ceildivpow2(tlprcystart, 1);
			brcbgxend = int_ceildivpow2(brprcxend, 1);
			brcbgyend = int_ceildivpow2(brprcyend, 1);
			cbgwidthexpn = pdx - 1;
			cbgheightexpn = pdy - 1;
		}
			
		cblkwidthexpn = int_min(info_out->tccpcblkw, cbgwidthexpn);
		cblkheightexpn = int_min(info_out->tccpcblkh, cbgheightexpn);

		int resnomax = resno ? NUM_BANDS : 1;
		for (bandno = 0; bandno < resnomax; bandno++) {
			int x0b, y0b;
			int gain, numbps;
			int bbandno = resno == 0 ? 0 : bandno + 1;
				
			x0b = (bbandno == 1) || (bbandno == 3) ? 1 : 0;
			y0b = (bbandno == 2) || (bbandno == 3) ? 1 : 0;
				
			int bandx0;
			int bandy0;
			int bandx1;
			int bandy1;
						
			if (bbandno == 0) {
				// band border
				bandx0 = int_ceildivpow2(tx0, levelno);
				bandy0 = int_ceildivpow2(ty0, levelno);
				bandx1 = int_ceildivpow2(tx1, levelno);
				bandy1 = int_ceildivpow2(ty1, levelno);
				
			} else {
				bandx0 = int_ceildivpow2(tx0 - (1 << levelno) * x0b, levelno + 1);
				bandy0 = int_ceildivpow2(ty0 - (1 << levelno) * y0b, levelno + 1);
				bandx1 = int_ceildivpow2(tx1 - (1 << levelno) * x0b, levelno + 1);
				bandy1 = int_ceildivpow2(ty1 - (1 << levelno) * y0b, levelno + 1);
			}
				
			gain = dwt_getgain(bbandno);
			numbps = PRECISION + gain;
			double step = (float)((1.0 + info_out->mant[resno == 0 ? 0 : 3 * (resno - 1) + bandno + 1]/ 2048.0) * pow(2.0, numbps - info_out->expn[resno == 0 ? 0 : 3 * (resno - 1) + bandno + 1]));
			info_out->stepsize[resno][bandno] = step;
			info_out->numbpsb[resno][bandno] = info_out->expn[resno == 0 ? 0 : 3 * (resno - 1) + bandno + 1] + NUMGBITS - 1;
				
			for (precno = 0; precno < NUM_PRECS; precno++) {
				int tlcblkxstart, tlcblkystart, brcblkxend, brcblkyend;

				int cbgxstart = tlcbgxstart + (precno % rpw) * (1 << cbgwidthexpn);
				int cbgystart = tlcbgystart + (precno / rpw) * (1 << cbgheightexpn);
				int cbgxend = cbgxstart + (1 << cbgwidthexpn);
				int cbgyend = cbgystart + (1 << cbgheightexpn);
				
				// precinct size (global)
				int precx0 = int_max(cbgxstart, bandx0);
				int precy0 = int_max(cbgxstart, bandy0);
				int precx1 = int_min(cbgxend, bandx1);
				int precy1 = int_min(cbgyend, bandy1);
				
				int px0 = int_max(cbgxstart,bandx0);
				int py0 = int_max(cbgystart,bandy0);
				int px1 = int_min(cbgxend,bandx1);
				int py1 = int_min(cbgyend,bandy1);

				tlcblkxstart = int_floordivpow2(px0, cblkwidthexpn) << cblkwidthexpn;
				tlcblkystart = int_floordivpow2(py0, cblkheightexpn) << cblkheightexpn;
				brcblkxend = int_ceildivpow2(px1, cblkwidthexpn) << cblkwidthexpn;
				brcblkyend = int_ceildivpow2(py1, cblkheightexpn) << cblkheightexpn;
				info_out->cw[resno][bandno][precno] = (brcblkxend - tlcblkxstart) >> cblkwidthexpn;
				info_out->ch[resno][bandno][precno] = (brcblkyend - tlcblkystart) >> cblkheightexpn;
				info_out->codebs[resno][bandno][precno] = ((brcblkxend - tlcblkxstart) >> cblkwidthexpn)*((brcblkyend - tlcblkystart) >> cblkheightexpn);

				for (cblkno = 0; cblkno < info_out->codebs[resno][bandno][precno]; cblkno++) {
					int cblkxstart = tlcblkxstart + (cblkno % info_out->cw[resno][bandno][precno]) * (1 << cblkwidthexpn);
					int cblkystart = tlcblkystart + (cblkno / info_out->cw[resno][bandno][precno]) * (1 << cblkheightexpn);
					int cblkxend = cblkxstart + (1 << cblkwidthexpn);
					int cblkyend = cblkystart + (1 << cblkheightexpn);

					info_out->codebx0[resno][bandno][precno][cblkno] = int_max(cblkxstart, precx0);
					info_out->codeby0[resno][bandno][precno][cblkno] = int_max(cblkystart, precy0);
					info_out->codebx1[resno][bandno][precno][cblkno] = int_min(cblkxend, precx1);
					info_out->codeby1[resno][bandno][precno][cblkno] = int_min(cblkyend, precy1);
				}
			}
		}
	}
}

int dwt_getgain(int orient) {
	if (orient == 0)
		return 0;
	if (orient == 1 || orient == 2)
		return 1;
	return 2;
}

int int_min(int a, int b) {
	return a < b ? a : b;
}

int int_ceildiv(int a, int b) {
	return (a + b - 1) / b;
}

int int_ceildivpow2(int a, int b) {
	return (a + (1 << b) - 1) >> b;
}

int int_floordivpow2(int a, int b) {
	return a >> b;
}

void j2k_setup_encoder( infos_t * info_out) {
	int j;

	info_out->tccpcblkw = int_floorlog2(CBLKW);
	info_out->tccpcblkh = int_floorlog2(CBLKH);

	for (j = 0; j < NUM_RES; j++) {
		info_out->prcw[j] = 15;
		info_out->prch[j] = 15;
	}
	dwt_calc_explicit_stepsizes(PRECISION,info_out);
}

int int_floorlog2(int a) {
	int l;
	for (l = 0; a > 1; l++) {
		a >>= 1;
	}
	return l;
}

void dwt_encode_stepsize(int stepsize, int numbps, infos_t * info, int bandno) {
	int p, n;
	p = int_floorlog2(stepsize) - 13;
	n = 11 - int_floorlog2(stepsize);
	info->mant[bandno] = (n < 0 ? stepsize >> -n : stepsize << n) & 0x7ff;
	info->expn[bandno] = numbps - p;
}

static const double dwt_norms_real[4][10] = {
	{1.000, 1.965, 4.177, 8.403, 16.90, 33.84, 67.69, 135.3, 270.6, 540.9},
	{2.022, 3.989, 8.355, 17.04, 34.27, 68.63, 137.3, 274.6, 549.0},
	{2.022, 3.989, 8.355, 17.04, 34.27, 68.63, 137.3, 274.6, 549.0},
	{2.080, 3.865, 8.307, 17.18, 34.71, 69.59, 139.3, 278.6, 557.2}
};

void dwt_calc_explicit_stepsizes( int prec,infos_t * info) {
	int numbands, bandno;
	
	/*MA-a: TO_FIX*/
	numbands = 3 * NUM_RES - 2;
		
	for (bandno = 0; bandno < numbands; bandno++) {
		double stepsize;
		int level, orient, gain, resno;
		resno = (bandno == 0) ? 0 : ((bandno - 1) / 3 + 1);
		orient = (bandno == 0) ? 0 : ((bandno - 1) % 3 + 1);
		level = NUM_RES - 1 - resno;
		gain = (QMFBID == 0) ? 0 : ((orient == 0) ? 0 : (((orient == 1) || (orient == 2)) ? 1 : 2));
		
		if (QNTSTY == J2K_CCP_QNTSTY_NOQNT) {
			stepsize = 1.0;
		} else {
			double norm = dwt_norms_real[orient][level];
			stepsize = (1 << (gain)) / norm;
		}
		dwt_encode_stepsize((int) floor(stepsize * 8192.0), prec + gain,info, bandno);
	}
}

