/*******************************************************************\

Copyright (c) 2008 Leiden University (LERC group at LIACS).
All rights reserved.

The use and distribution terms for this software are covered by the 
Common Public License 1.0 (http://opensource.org/licenses/cpl1.0.txt)
which can be found in the file LICENSE at the root of this distribution.
By using this software in any fashion, you are agreeing to be bound by 
the terms of this license.

You must not remove this notice, or any other, from this software.

\*******************************************************************/

#ifndef T1_H_
#define T1_H_

#include "jpeg2000_func.h"

#define MQC_NUMCTXS 32
#define T1_NMSEDEC_BITS 7
#define T1_NMSEDEC_FRACBITS (T1_NMSEDEC_BITS-1)
#define T1_TYPE_MQ 0
#define T1_TYPE_RAW 1
#define T1_NUMCTXS_AGG 1
#define T1_NUMCTXS_MAG 3
#define T1_NUMCTXS_ZC 9
#define T1_NUMCTXS_SC 5
#define T1_CTXNO_ZC 0
#define T1_CTXNO_SC (T1_CTXNO_ZC+T1_NUMCTXS_ZC)
#define T1_CTXNO_MAG (T1_CTXNO_SC+T1_NUMCTXS_SC)
#define T1_CTXNO_AGG (T1_CTXNO_MAG+T1_NUMCTXS_MAG)
#define T1_CTXNO_UNI (T1_CTXNO_AGG+T1_NUMCTXS_AGG)
#define J2K_CCP_CBLKSTY_SEGSYM 0x20
#define J2K_CCP_CBLKSTY_TERMALL 0x04
#define J2K_CCP_CBLKSTY_LAZY 0x01
#define J2K_CCP_CBLKSTY_RESET 0x02
#define J2K_CCP_CBLKSTY_PTERM 0x10
#define J2K_CCP_CBLKSTY_VSC 0x08
#define T1_SIG_NE 0x0001	// Context orientation : North-East direction
#define T1_SIG_SE 0x0002	// Context orientation : South-East direction
#define T1_SIG_SW 0x0004	// Context orientation : South-West direction
#define T1_SIG_NW 0x0008	// Context orientation : North-West direction
#define T1_SIG_N 0x0010		// Context orientation : North direction
#define T1_SIG_E 0x0020		// Context orientation : East direction
#define T1_SIG_S 0x0040		// Context orientation : South direction
#define T1_SIG_W 0x0080		// Context orientation : West direction
#define T1_SIG_OTH (T1_SIG_N|T1_SIG_NE|T1_SIG_E|T1_SIG_SE|T1_SIG_S|T1_SIG_SW|T1_SIG_W|T1_SIG_NW)
#define T1_SIG_PRIM (T1_SIG_N|T1_SIG_E|T1_SIG_S|T1_SIG_W)
#define T1_SGN_N 0x0100
#define T1_SGN_E 0x0200
#define T1_SGN_S 0x0400
#define T1_SGN_W 0x0800
#define T1_SGN (T1_SGN_N|T1_SGN_E|T1_SGN_S|T1_SGN_W)
#define T1_SIG 0x1000
#define T1_VISIT 0x4000
#define T1_REFINE 0x2000
#define J2K_CCP_CSTY_PRT 0x01
#define J2K_CCP_QNTSTY_NOQNT 0
#define OPJ_PATH_LEN 4096
#define J2K_CCP_QNTSTY_SEQNT 2
#define J2K_CCP_CSTY_PRT 0x01

#define DATASIZE WIDTH_TILE * HEIGHT_TILE
#define FLAGSIZE (WIDTH_TILE + 2) * (HEIGHT_TILE + 2)

typedef struct infos_t {
	int bandx0[NUM_RES][NUM_BANDS];
	int bandy0[NUM_RES][NUM_BANDS];
	int precx0[NUM_RES][NUM_BANDS][NUM_PRECS];
	int precy0[NUM_RES][NUM_BANDS][NUM_PRECS];
	int precx1[NUM_RES][NUM_BANDS][NUM_PRECS];
	int precy1[NUM_RES][NUM_BANDS][NUM_PRECS];
	int codebs[NUM_RES][NUM_BANDS][NUM_PRECS];
	int codebx0[NUM_RES][NUM_BANDS][NUM_PRECS][NUM_CODEBS];
	int codeby0[NUM_RES][NUM_BANDS][NUM_PRECS][NUM_CODEBS];
	int codebx1[NUM_RES][NUM_BANDS][NUM_PRECS][NUM_CODEBS];
	int codeby1[NUM_RES][NUM_BANDS][NUM_PRECS][NUM_CODEBS];
	int numbpsb[NUM_RES][NUM_BANDS];
	double stepsize [NUM_RES][NUM_BANDS];
	int prcw[NUM_RES];
	int prch[NUM_RES];
	int tx0,ty0,tx1,ty1;
	int tccpcblkw;
	int tccpcblkh;
	int expn[(NUM_RES * 3) - 2];
	int mant[(NUM_RES * 3) - 2];
	int num;
} infos_t;

typedef struct opj_mqc_state {
	// the probability of the Least Probable Symbol (0.75->0x8000, 1.5->0xffff)
	unsigned int qeval;
	// the Most Probable Symbol (0 or 1) 
	int mps;
	// next state if the next encoded symbol is the MPS
	struct opj_mqc_state *nmps;
	// next state if the next encoded symbol is the LPS
	struct opj_mqc_state *nlps;
} opj_mqc_state_t;

typedef struct opj_mqc {
	unsigned int c;
	unsigned int a;
	unsigned int ct;
	unsigned char *bp;
	unsigned char *start;
	unsigned char *end;
	opj_mqc_state_t *ctxs[MQC_NUMCTXS];
	opj_mqc_state_t **curctx;
} opj_mqc_t;

typedef struct opj_raw {
	// temporary buffer where bits are coded or decoded
	unsigned char c;
	// number of bits already read or free to write
	unsigned int ct;
	// maximum length to decode
	unsigned int lenmax;
	// length decoded
	unsigned int len;
	// pointer to the current position in the buffer
	unsigned char *bp;
	// pointer to the start of the buffer
	unsigned char *start;
	// pointer to the end of the buffer
	unsigned char *end;
} opj_raw_t;

typedef unsigned short flag_t;

typedef struct opj_t1 {
	// MQC component
	opj_mqc_t mqc;
	// RAW component
	opj_raw_t raw;

	int data[DATASIZE];
	flag_t flags[FLAGSIZE];
	int w;
	int h;
	int datasize;
	int flagssize;
	int flags_stride;
} opj_t1_t;

#define MACRO_t1_flags(x,y) t1->flags[((x)*(t1->flags_stride))+(y)]

static void t1_encode_cblks(opj_t1_t *t1, const tile_c_t * t, codeblock_t * block_out, infos_t * info, passes_t * passes_out);
static void allocate_buffers(opj_t1_t *t1,	int w, int h);
static void t1_encode_cblk(opj_t1_t *t1, pass_t * pass, int orient,int compno,int level, int qmfbid, double stepsize,int cblksty,int numcomps, unsigned char * data, int * numbps, int * passnum, double * distotile);
static int int_abs(int a);
static void mqc_resetstates(opj_mqc_t *mqc);
static void mqc_setstate(opj_mqc_t *mqc, int ctxno, int msb, int prob);
static void mqc_init_enc(opj_mqc_t *mqc, unsigned char *bp);
static void mqc_setcurctx(opj_mqc_t *mqc, int ctxno);
static void t1_enc_sigpass(opj_t1_t *t1,int bpno,int orient,int *nmsedec,char type,int cblksty);
static void t1_enc_sigpass_step(opj_t1_t *t1,flag_t *flagsp,int *datap,int orient,int bpno,int one,int *nmsedec,char type,	int vsc);
static char t1_getctxno_zc(int f, int orient);
static void mqc_bypass_enc(opj_mqc_t *mqc, int d);
static void mqc_encode(opj_mqc_t *mqc, int d);
static void mqc_codemps(opj_mqc_t *mqc);
static void mqc_renorme(opj_mqc_t *mqc);
static void mqc_byteout(opj_mqc_t *mqc);
static void mqc_codelps(opj_mqc_t *mqc);
static short t1_getnmsedec_sig(int x, int bitpos);
static char t1_getctxno_sc(int f);
static char t1_getspb(int f);
static void t1_updateflags(flag_t *flagsp, int s, int stride);
static void t1_enc_refpass_step(opj_t1_t *t1,flag_t *flagsp,int *datap,int bpno,int one,int *nmsedec,char type,int vsc);
static short t1_getnmsedec_ref(int x, int bitpos);
static char t1_getctxno_mag(int f);
static void t1_enc_refpass(opj_t1_t *t1,int bpno,int *nmsedec,char type,int cblksty);
static void t1_enc_clnpass(opj_t1_t *t1,int bpno,int orient,int *nmsedec,int cblksty);
static void t1_enc_clnpass_step(opj_t1_t *t1,flag_t *flagsp,int *datap,int orient,int bpno,int one,int *nmsedec,int partial,int vsc);
static void mqc_segmark_enc(opj_mqc_t *mqc);
static double t1_getwmsedec(int nmsedec,int compno,int level,int orient,int bpno,int qmfbid,double stepsize,int numcomps);
static double mct_getnorm(int compno);
static double dwt_getnorm(int level, int orient);
static void mqc_flush(opj_mqc_t *mqc);
static void mqc_setbits(opj_mqc_t *mqc);
static void mqc_bypass_init_enc(opj_mqc_t *mqc);
static void mqc_restart_init_enc(opj_mqc_t *mqc);
static int mqc_numbytes(opj_mqc_t *mqc);
static void mqc_reset_enc(opj_mqc_t *mqc);
static void mqc_erterm_enc(opj_mqc_t *mqc);
static int int_max(int a, int b);
static int int_floorlog2(int a);
static void tcd_init_encode(infos_t * info_out, int num);
static int dwt_getgain(int orient);
static int int_min(int a, int b);
static int int_ceildiv(int a, int b);
static int int_ceildivpow2(int a, int b);
static int int_floordivpow2(int a, int b);
static void dwt_calc_explicit_stepsizes(int prec,infos_t * info);
static void dwt_encode_stepsize(int stepsize, int numbps,infos_t * info, int bandno);
static void j2k_setup_encoder(infos_t * info_out);

#endif /*T1_H_*/

