/*******************************************************************\

Copyright (c) 2008 Leiden University (LERC group at LIACS).
All rights reserved.

The use and distribution terms for this software are covered by the 
Common Public License 1.0 (http://opensource.org/licenses/cpl1.0.txt)
which can be found in the file LICENSE at the root of this distribution.
By using this software in any fashion, you are agreeing to be bound by 
the terms of this license.

You must not remove this notice, or any other, from this software.

\*******************************************************************/

#ifndef T2_H_
#define T2_H_

#include "jpeg2000_func.h"
#ifndef PC
#include "MemoryMap.h"
#endif

#define DBL_MAX		__DBL_MAX__
#define J2K_CCP_QNTSTY_NOQNT 0
#define J2K_CCP_CSTY_PRT 0x01
#define J2K_CCP_QNTSTY_SIQNT 1
#define OPJ_STREAM_WRITE 0x0002
#define J2K_CP_CSTY_SOP 0x02
#define J2K_CP_CSTY_EPH 0x04

typedef enum T2_MODE {
	THRESH_CALC = 0,	// Function called in Rate allocation process
	FINAL_PASS = 1		// Function called in Tier 2 process
}J2K_T2_MODE;

#define NUM_LAYERS 1
#define INCLUDESIZE NUM_RES * 6

typedef struct infos_t {
	int codebs[NUM_RES][NUM_BANDS][NUM_PRECS];
	int codebx0[NUM_RES][NUM_BANDS][NUM_PRECS][NUM_CODEBS];
	int codeby0[NUM_RES][NUM_BANDS][NUM_PRECS][NUM_CODEBS];
	int codebx1[NUM_RES][NUM_BANDS][NUM_PRECS][NUM_CODEBS];
	int codeby1[NUM_RES][NUM_BANDS][NUM_PRECS][NUM_CODEBS];
	int cw[NUM_RES][NUM_BANDS][NUM_PRECS];
	int ch[NUM_RES][NUM_BANDS][NUM_PRECS];
	int numbpsb[NUM_RES][NUM_BANDS];
	double stepsize[NUM_RES][NUM_BANDS];
	int prcw[NUM_RES];
	int prch[NUM_RES];
	int tx0,ty0,tx1,ty1;
	int tccpcblkw;
	int tccpcblkh;
	int expn[(3 * NUM_BANDS) - 2];
	int mant[(3 * NUM_BANDS) - 2];
	int num;
	int compnum;
} infos_t;

typedef struct opj_bio {
	// pointer to the start of the buffer
	unsigned char *start;
	// pointer to the end of the buffer
	unsigned char *end;
	// pointer to the present position in the buffer
	unsigned char *bp;
	// temporary place where each byte is read or written
	unsigned int buf;
	// coder : number of bits free to write. decoder : number of bits read
	int ct;
} opj_bio_t;

typedef enum PROG_ORDER {
	PROG_UNKNOWN = -1,	// place-holder
	LRCP = 0,		// layer-resolution-component-precinct order
	RLCP = 1,		// resolution-layer-component-precinct order
	RPCL = 2,		// resolution-precinct-component-layer order
	PCRL = 3,		// precinct-component-resolution-layer order
	CPRL = 4		// component-precinct-resolution-layer order
} OPJ_PROG_ORDER;

typedef struct j2k_prog_order{
	OPJ_PROG_ORDER enum_prog;
	char str_prog[5];
}j2k_prog_order_t;

typedef struct layer_t{
  int numpasses;
  int len;
  float disto;
} layer_t;

typedef struct layers_t{
	layer_t layer[NUM_RES][NUM_BANDS][NUM_PRECS][NUM_CODEBS][NUM_LAYERS];
	int numpasses[NUM_RES][NUM_BANDS][NUM_PRECS][NUM_CODEBS];
	double distolayer[NUM_LAYERS];
	int nbpixc;
	int nbpix;
	float rates[NUM_LAYERS];
	float distoratio[NUM_LAYERS];
} layers_t;

typedef struct res_t{
	int pdx[NUM_RES];
	int pdy[NUM_RES];
	int pw[NUM_RES];
	int ph[NUM_RES];
} res_t;

typedef struct node {
  struct node *parent;
  int value;
  int low;
  int known;
} node_t;

typedef struct tree_t {
  int numleafsh;
  int numleafsv;
  int numnodes;
  node_t nodes[NUM_CODEBS];
} tree_t;

typedef struct opj_tgt_node {
  struct opj_tgt_node *parent;
  int value;
  int low;
  int known;
} opj_tgt_node_t;

typedef struct opj_tgt_tree {
  int numleafsh;
  int numleafsv;
  int numnodes;
  opj_tgt_node_t nodes[1];
} opj_tgt_tree_t;

typedef struct tree_comp_t{
	opj_tgt_tree_t incltree[NUM_RES][NUM_BANDS][NUM_PRECS];
  	opj_tgt_tree_t imsbtree[NUM_RES][NUM_BANDS][NUM_PRECS];
}tree_comp_t;

typedef struct opj_poc {
	// Resolution num start, Component num start, given by POC
	int resno0, compno0;
	// Layer num end,Resolution num end, Component num end, given by POC
	int layno1, resno1, compno1;
	// Layer num start,Precinct num start, Precinct num end
	int layno0, precno0, precno1;
	// Progression order enum
	OPJ_PROG_ORDER prg1,prg;
	// Progression order string
	char progorder[5];
	// Tile number 
	int tile;
	// Start and end values for Tile width and height
	int tx0,tx1,ty0,ty1;
	// Start value, initialised in pi_initialise_encode
	int layS, resS, compS, prcS;
	// End value, initialised in pi_initialise_encode
	int layE, resE, compE, prcE;
	// Start and end values of Tile width and height, initialised in pi_initialise_encode
	int txS,txE,tyS,tyE,dx,dy;
	// Temporary values for Tile parts, initialised in pi_create_encode
	int lay_t, res_t, comp_t, prc_t,tx0_t,ty0_t;
} opj_poc_t;

typedef enum CINEMA_MODE {
	OFF = 0,			// Not Digital Cinema
	CINEMA2K_24 = 1,	// 2K Digital Cinema at 24 fps
	CINEMA2K_48 = 2,	// 2K Digital Cinema at 48 fps
	CINEMA4K_24 = 3		// 4K Digital Cinema at 24 fps
}OPJ_CINEMA_MODE;

typedef enum RSIZ_CAPABILITIES {
	STD_RSIZ = 0,		// Standard JPEG2000 profile
	CINEMA2K = 3,		// Profile name for a 2K image
	CINEMA4K = 4		// Profile name for a 4K image
} OPJ_RSIZ_CAPABILITIES;

typedef struct opj_pi_resolution {
  int pdx, pdy;
  int pw, ph;
} opj_pi_resolution_t;

typedef struct opj_pi_comp {
  int dx, dy;
  int numresolutions;
  opj_pi_resolution_t resolutions[NUM_RES];
} opj_pi_comp_t;

typedef struct opj_pi_iterator {
	// Enabling Tile part generation
	char tp_on;
	// precise if the packet has been already used (usefull for progression order change)
	short int include[INCLUDESIZE];
	// layer step used to localize the packet in the include vector
	int step_l;
	// resolution step used to localize the packet in the include vector 
	int step_r;	
	// component step used to localize the packet in the include vector
	int step_c;	
	// precinct step used to localize the packet in the include vector 
	int step_p;	
	// component that identify the packet
	int compno;
	// resolution that identify the packet
	int resno;
	// precinct that identify the packet
	int precno;
	// layer that identify the packet
	int layno;   
	// 0 if the first packet
	int first;
	// progression order change information
	opj_poc_t poc;
	// number of components in the image
	int numcomps;
	// Components
	opj_pi_comp_t comps;
	int tx0, ty0, tx1, ty1;
	int x, y, dx, dy;
	// Temporary values for Tile parts, initialised in pi_create_encode
	int lay_t, res_t, comp_t, prc_t,tx0_t,ty0_t;
} opj_pi_iterator_t;

static int t2_encode_packets(int tileno, int maxlayers, unsigned char *dest, int len, int tpnum, int tppos,int pino, J2K_T2_MODE t2_mode, infos_t * info,const codeblock_t * blocks_in, layers_t * lay,res_t * rest, tree_comp_t * tree, const passes_t * passes_in);
static int t2_encode_packet(opj_pi_iterator_t *pi, unsigned char *dest, int length, int tileno, infos_t * info,const codeblock_t * blocks_in, layers_t * lay, int * num, tree_comp_t * tree, const passes_t * passes_in);
static int tcd_rateallocate(infos_t * info, const codeblock_t * blocks_in, layers_t * lay,tree_comp_t * tree, const passes_t * passes_in);
static void tcd_makelayer(int layno, double thresh, int final, const infos_t * info, layers_t * lay, const passes_t * passes_in);
static void pi_create_encode( opj_pi_iterator_t *pi, int tileno, int pino,int tpnum, int tppos, const infos_t * info);
static int pi_check_next_level(int pos, opj_pi_iterator_t *pi,int tileno, int pino, char *prog,const infos_t * info);
static int pi_next_lrcp(opj_pi_iterator_t * pi);
static int pi_next_rlcp(opj_pi_iterator_t * pi);
static int pi_next_rpcl(opj_pi_iterator_t * pi,res_t * res);
static int pi_next_pcrl(opj_pi_iterator_t * pi,res_t * res);
static int pi_next_cprl(opj_pi_iterator_t * pi,res_t * rest);
static void tgt_setvalue(opj_tgt_tree_t *tree, int leafno, int value);
static void bio_init_enc(opj_bio_t *bio, unsigned char *bp, int len);
static void bio_write(opj_bio_t *bio, int v, int n);
static void tgt_encode(opj_bio_t *bio, opj_tgt_tree_t *tree, int leafno, int threshold);
static void t2_putnumpasses(opj_bio_t *bio, int n);
static void t2_putcommacode(opj_bio_t *bio, int n);
static int bio_flush(opj_bio_t *bio);
static int bio_byteout(opj_bio_t *bio);
static int bio_numbytes(opj_bio_t *bio);
static void bio_putbit(opj_bio_t *bio, int b);
static int int_min(int a, int b);
static int int_max(int a, int b);
static int int_ceildiv(int a, int b);
static int int_ceildivpow2(int a, int b);
static int int_floordivpow2(int a, int b);
static void tgt_create(opj_tgt_tree_t *tree, int numleafsh, int numleafsv);
static void tgt_reset(opj_tgt_tree_t *tree);
static char *j2k_convert_progression_order(OPJ_PROG_ORDER prg_order);
static int int_floorlog2(int a);
static int pi_next(opj_pi_iterator_t * pi,res_t * rest);
static void tcd_init_encode(infos_t * info_out, int num);
static int dwt_getgain(int orient);
static void dwt_calc_explicit_stepsizes(int prec,infos_t * info);
static void dwt_encode_stepsize(int stepsize, int numbps,infos_t * info,int bandno);
static void j2k_setup_encoder(infos_t * info_out);
static void pi_initialise_encode(opj_pi_iterator_t *pi, int tileno, J2K_T2_MODE t2_mode, const infos_t * info, res_t * rest);

#endif /*T2_H_*/

