/*******************************************************************\

Copyright (c) 2008 Leiden University (LERC group at LIACS).
All rights reserved.

The use and distribution terms for this software are covered by the 
Common Public License 1.0 (http://opensource.org/licenses/cpl1.0.txt)
which can be found in the file LICENSE at the root of this distribution.
By using this software in any fashion, you are agreeing to be bound by 
the terms of this license.

You must not remove this notice, or any other, from this software.

\*******************************************************************/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "VideoOut.h"

void mainVideoOut(const packet_t * packets_in)
{
	opj_cio_t cio;	//Codestream
	int compno,tileno,totlen,ct;
	int n;
	long int k;
	infos_t info;
	infos_t * info_in;
	info_in = &info;
	
	j2k_setup_encoder(info_in);
	tcd_init_encode(info_in,packets_in->num);

	tileno = packets_in->num;
	compno = packets_in->numcomp;
	
	cio.buffer = (int *) &outMemory[offset];
	cio.start = cio.buffer;
	cio.end = cio.buffer + (int) (IMAGE_SIZE / NUMTILES);
	cio.bp = cio.buffer;

	if ((tileno == 0) && (compno == 0))
	{		
		j2k_write_soc(&cio);
		j2k_write_siz(&cio,info_in);
		j2k_write_cod(&cio,info_in);
		j2k_write_qcd(&cio,info_in);
	}
	

	if (compno == 0) j2k_write_sot(&cio,packets_in);

	if((CUR_TP_NUM == 0) && (compno == 0)){
		for (ct = 1; ct < NUM_COMPS; ct++)
		{
			j2k_write_coc(&cio, ct,info_in);
			j2k_write_qcc(&cio, ct,info_in);
		}
	}

	if (compno == 0) cio_write(&cio, J2K_MS_SOD, 2);
	
	if ((tileno == 0) && (compno == 0)) {
		cio.sod_start = cio_tell(&cio);
	}

	int c;
	
	for (c = 0; c < packets_in->lenght; c ++)
	{
		*(cio.bp ++) = packets_in->data[c];
	}
	
	// Writing Psot in SOT marker
	if (compno == 2)
	{
		totlen = (cio_tell(&cio) + t_offset) - t_start;
		cio_seek(&cio, t_start + 6 - t_offset);
		cio_write(&cio, totlen, 4);
		cio_seek(&cio, t_start + totlen - t_offset);
	}
	
	
	// Write file if last tile
	if ((tileno == NUMTILES -1) && (compno == 2)){
		j2k_write_eoc(&cio);
	}
	
	if (compno == 2)
	{
		t_offset = 0;
	}
	else
	{
		t_offset += cio_tell(&cio);
	}
	
	n = cio_tell(&cio);
	offset = offset + n;

	if ((tileno == NUMTILES -1) && (compno == 2)){	
#ifdef PC
		f = fopen("test.j2k","wb");

		for(k = 0; k < offset; k++)
		{
			putc(outMemory[k],f);
		}	
		fclose(f);	
		
#else
		*ZBT_MEM = offset;
#endif
	}
}

void j2k_write_eoc(opj_cio_t *cio) {
	cio_write(cio, J2K_MS_EOC, 2);
}

int cio_byteout(opj_cio_t *cio, unsigned char v) {
	if (cio->bp >= cio->end) {
		return false;
	}
	*cio->bp++ = v;
	return true;
}

unsigned int cio_write(opj_cio_t *cio, unsigned int v, int n) {
	int i;
	for (i = n - 1; i >= 0; i--) {
		if( !cio_byteout(cio, (unsigned char) ((v >> (i << 3)) & 0xff)) )
			return 0;
	}
	return n;
}

int cio_tell(opj_cio_t *cio) {
	return cio->bp - cio->start;
}

void cio_skip(opj_cio_t *cio, int n) {
	cio->bp += n;
}

void cio_seek(opj_cio_t *cio, int pos) {
	cio->bp = cio->start + pos;
}

void j2k_write_soc(opj_cio_t *cio) {
	cio_write(cio, J2K_MS_SOC, 2);
}

void j2k_write_siz(opj_cio_t *cio,const infos_t * info) {
	int i;
	int lenp, len;

	cio_write(cio, J2K_MS_SIZ, 2);
	lenp = cio_tell(cio);
	cio_skip(cio, 2);
	cio_write(cio, STD_RSIZ, 2);
	cio_write(cio, WIDTH_IMAGE, 4);
	cio_write(cio, HEIGHT_IMAGE, 4);
	cio_write(cio, OFFSET_X, 4);
	cio_write(cio, OFFSET_Y, 4);
	cio_write(cio, WIDTH_TILE, 4);
	cio_write(cio, HEIGHT_TILE, 4);
	cio_write(cio, info->tx0, 4);
	cio_write(cio, info->ty0, 4);
	cio_write(cio, NUM_COMPS, 2);
	for (i = 0; i < NUM_COMPS; i++) {
		cio_write(cio,PRECISION - 1 + (SIGN << 7), 1);
		cio_write(cio,SUBSAMPLING_DX, 1);
		cio_write(cio,SUBSAMPLING_DY, 1);
	}
	len = cio_tell(cio) - lenp;
	cio_seek(cio, lenp);
	cio_write(cio, len, 2);	
	cio_seek(cio, lenp + len);
}

void j2k_write_cox(opj_cio_t *cio, int compno,const infos_t * info) {
	int i;
	
	cio_write(cio, NUM_RES - 1, 1);	
	cio_write(cio, info->tccpcblkw - 2, 1);
	cio_write(cio, info->tccpcblkh - 2, 1);
	cio_write(cio, MODE, 1);
	cio_write(cio, QMFBID, 1);
	
	if (CSTY & J2K_CCP_CSTY_PRT) {
		for (i = 0; i < NUM_RES; i++) {
			cio_write(cio, info->prcw[i][0] + (info->prch[i][0] << 4), 1);
		}
	}
}

void j2k_write_cod(opj_cio_t *cio, const infos_t * info) {

	int lenp, len;

	cio_write(cio, J2K_MS_COD, 2);
	
	lenp = cio_tell(cio);
	cio_skip(cio, 2);
	
	cio_write(cio, CSTY, 1);
	cio_write(cio, PROG_ORDER, 1);
	cio_write(cio, NUM_LAYERS, 2);
	cio_write(cio, MCT, 1);
	
	j2k_write_cox(cio, 0,info);
	len = cio_tell(cio) - lenp;
	cio_seek(cio, lenp);
	cio_write(cio, len, 2);	
	cio_seek(cio, lenp + len);
}

void j2k_write_qcd(opj_cio_t *cio,const infos_t * info) {
	int lenp, len;

	cio_write(cio, J2K_MS_QCD, 2);
	lenp = cio_tell(cio);
	cio_skip(cio, 2);
	j2k_write_qcx(cio, 0,info);
	len = cio_tell(cio) - lenp;
	cio_seek(cio, lenp);
	cio_write(cio, len, 2);
	cio_seek(cio, lenp + len);
}

void j2k_write_qcx(opj_cio_t *cio, int compno,const infos_t * info) {
	int bandno, numbands;
	int expn, mant;
	
	cio_write(cio, QNTSTY + (NUMGBITS << 5), 1);
	numbands = QNTSTY == J2K_CCP_QNTSTY_SIQNT ? 1 : NUM_RES * 3 - 2;
	
	for (bandno = 0; bandno < numbands; bandno++) {
		
		expn = info->expn[bandno];
		mant = info->mant[bandno];
		
		if (QNTSTY == J2K_CCP_QNTSTY_NOQNT) {
			cio_write(cio, expn << 3, 1);
		} else {
			cio_write(cio, (expn << 11) + mant, 2);
		}
	}
}

void j2k_write_coc(opj_cio_t *cio, int compno, const infos_t * info) {
	int lenp, len;

	cio_write(cio, J2K_MS_COC, 2);
	lenp = cio_tell(cio);
	cio_skip(cio, 2);
	cio_write(cio, compno, NUM_COMPS <= 256 ? 1 : 2);
	cio_write(cio, CSTY, 1);
	j2k_write_cox(cio, compno,info);
	len = cio_tell(cio) - lenp;
	cio_seek(cio, lenp);
	cio_write(cio, len, 2);
	cio_seek(cio, lenp + len);
}

void j2k_write_qcc(opj_cio_t *cio, int compno,const infos_t * info) {
	int lenp, len;

	cio_write(cio, J2K_MS_QCC, 2);
	lenp = cio_tell(cio);
	cio_skip(cio, 2);
	cio_write(cio, compno, NUM_COMPS <= 256 ? 1 : 2);
	j2k_write_qcx(cio, compno,info);
	len = cio_tell(cio) - lenp;
	cio_seek(cio, lenp);
	cio_write(cio, len, 2);
	cio_seek(cio, lenp + len);
}

void j2k_write_sot(opj_cio_t *cio, const packet_t * packets_in) {
	int lenp, len;

	cio->sot_start = cio_tell(cio);
	t_start = cio->sot_start;
	cio_write(cio, J2K_MS_SOT, 2);
	lenp = cio_tell(cio);
	cio_skip(cio, 2);
	cio_write(cio, packets_in->num, 2);
	cio_skip(cio, 4);
	cio_write(cio, CUR_TP_NUM, 1);
	cio_write(cio, CUR_TOTNUM_TP, 1);
	len = cio_tell(cio) - lenp;
	cio_seek(cio, lenp);
	cio_write(cio, len, 2);
	cio_seek(cio, lenp + len);
}

int int_max(int a, int b) {
	return (a > b) ? a : b;
}

void tcd_init_encode(infos_t * info_out,int num) {
	int p = num % HTILES;
	int q = num / HTILES;
	
	info_out->tx0 = int_max(0 + p * WIDTH_TILE, 0);
	info_out->ty0 = int_max(0 + q * HEIGHT_TILE, 0);
	info_out->tx1 = int_min(0 + (p + 1) * WIDTH_TILE, WIDTH_IMAGE);
	info_out->ty1 = int_min(0 + (q + 1) * HEIGHT_TILE, HEIGHT_IMAGE);
}

int int_min(int a, int b) {
	return a < b ? a : b;
}

void j2k_setup_encoder( infos_t * info_out) {
	int i, j;

	info_out->tccpcblkw = int_floorlog2(CBLKW);
	info_out->tccpcblkh = int_floorlog2(CBLKH);

	for (i = 0; i < NUM_COMPS; i++) {
		for (j = 0; j < NUM_RES; j++) {
			info_out->prcw[i][j] = 15;
			info_out->prch[i][j] = 15;
		}
		dwt_calc_explicit_stepsizes(PRECISION,info_out);
	}
}

int int_floorlog2(int a) {
	int l;
	for (l = 0; a > 1; l++) {
		a >>= 1;
	}
	return l;
}

void dwt_encode_stepsize(int stepsize, int numbps, infos_t * info, int bandno) {
	int p, n;
	p = int_floorlog2(stepsize) - 13;
	n = 11 - int_floorlog2(stepsize);
	info->mant[bandno] = (n < 0 ? stepsize >> -n : stepsize << n) & 0x7ff;
	info->expn[bandno] = numbps - p;
}

static const double dwt_norms_real[4][10] = {
	{1.000, 1.965, 4.177, 8.403, 16.90, 33.84, 67.69, 135.3, 270.6, 540.9},
	{2.022, 3.989, 8.355, 17.04, 34.27, 68.63, 137.3, 274.6, 549.0},
	{2.022, 3.989, 8.355, 17.04, 34.27, 68.63, 137.3, 274.6, 549.0},
	{2.080, 3.865, 8.307, 17.18, 34.71, 69.59, 139.3, 278.6, 557.2}
};

void dwt_calc_explicit_stepsizes( int prec,infos_t * info) {
	int numbands, bandno;
	
	/*MA-a: TO_FIX*/
	numbands = 3 * NUM_RES - 2;
		
	for (bandno = 0; bandno < numbands; bandno++) {
		double stepsize;
		int level, orient, gain, resno;
		resno = (bandno == 0) ? 0 : ((bandno - 1) / 3 + 1);
		orient = (bandno == 0) ? 0 : ((bandno - 1) % 3 + 1);
		level = NUM_RES - 1 - resno;
		gain = (QMFBID == 0) ? 0 : ((orient == 0) ? 0 : (((orient == 1) || (orient == 2)) ? 1 : 2));
		
		if (QNTSTY == J2K_CCP_QNTSTY_NOQNT) {
			stepsize = 1.0;
		} else {
			double norm = dwt_norms_real[orient][level];
			stepsize = (1 << (gain)) / norm;
		}
		dwt_encode_stepsize((int) floor(stepsize * 8192.0), prec + gain,info, bandno);
	}
}
