/*******************************************************************\

Copyright (c) 2008 Leiden University (LERC group at LIACS).
All rights reserved.

The use and distribution terms for this software are covered by the 
Common Public License 1.0 (http://opensource.org/licenses/cpl1.0.txt)
which can be found in the file LICENSE at the root of this distribution.
By using this software in any fashion, you are agreeing to be bound by 
the terms of this license.

You must not remove this notice, or any other, from this software.

\*******************************************************************/

#ifndef VIDEOOUT_H_
#define VIDEOOUT_H_

#include "jpeg2000_func.h"
#ifndef PC
#include "MemoryMap.h"
#endif

#define DBL_MAX		__DBL_MAX__
#define J2K_CCP_QNTSTY_NOQNT 0
#define J2K_CCP_CSTY_PRT 0x01
#define J2K_CCP_QNTSTY_SIQNT 1
#define OPJ_STREAM_WRITE 0x0002
#define J2K_CP_CSTY_SOP 0x02
#define J2K_CP_CSTY_EPH 0x04
#define J2K_MS_EOC 0xffd9	// EOC marker value
#define J2K_MS_SOC 0xff4f	// SOC marker value
#define J2K_MS_SIZ 0xff51	// SIZ marker value
#define J2K_MS_COD 0xff52	// COD marker value
#define J2K_MS_QCD 0xff5c	// QCD marker value
#define J2K_MS_COC 0xff53	// COC marker value
#define J2K_MS_QCC 0xff5d	// QCC marker value
#define J2K_MS_RGN 0xff5e	// RGN marker value
#define J2K_MS_TLM 0xff55	// TLM marker value
#define J2K_MS_POC 0xff5f	// POC marker value
#define J2K_MS_SOT 0xff90	// SOT marker value
#define J2K_MS_SOD 0xff93	// SOD marker value

typedef enum T2_MODE {
	THRESH_CALC = 0,	// Function called in Rate allocation process
	FINAL_PASS = 1		// Function called in Tier 2 process
}J2K_T2_MODE;

#define NUM_LAYERS 1
#define INCLUDESIZE 6

typedef struct infos_t {
	int prcw[NUM_COMPS][NUM_RES];
	int prch[NUM_COMPS][NUM_RES];
	int tx0,ty0,tx1,ty1;
	int tccpcblkw;
	int tccpcblkh;
	int expn[(3 * NUM_BANDS) - 2];
	int mant[(3 * NUM_BANDS) - 2];
} infos_t;

typedef enum PROG_ORDER {
	PROG_UNKNOWN = -1,	// place-holder
	LRCP = 0,		// layer-resolution-component-precinct order
	RLCP = 1,		// resolution-layer-component-precinct order
	RPCL = 2,		// resolution-precinct-component-layer order
	PCRL = 3,		// precinct-component-resolution-layer order
	CPRL = 4		// component-precinct-resolution-layer order
} OPJ_PROG_ORDER;

typedef struct j2k_prog_order{
	OPJ_PROG_ORDER enum_prog;
	char str_prog[5];
}j2k_prog_order_t;

typedef struct opj_cio {
	// pointer to the start of the buffer
	int *buffer;
	// pointer to the start of the stream
	int *start;
	// pointer to the end of the stream
	int *end;
	// pointer to the current position
	int *bp;
	int sot_start;
	int sod_start;
	int pos_correction;
	int tlm_start;
} opj_cio_t;

typedef enum CINEMA_MODE {
	OFF = 0,			// Not Digital Cinema
	CINEMA2K_24 = 1,	// 2K Digital Cinema at 24 fps
	CINEMA2K_48 = 2,	// 2K Digital Cinema at 48 fps
	CINEMA4K_24 = 3		// 4K Digital Cinema at 24 fps
}OPJ_CINEMA_MODE;

typedef enum RSIZ_CAPABILITIES {
	STD_RSIZ = 0,		// Standard JPEG2000 profile
	CINEMA2K = 3,		// Profile name for a 2K image
	CINEMA4K = 4		// Profile name for a 4K image
} OPJ_RSIZ_CAPABILITIES;

static long int offset = 0;
static int t_offset = 0;
static int t_start;

#ifdef PC
static FILE *f;
static int outMemory[IMAGE_SIZE];
#else
static volatile int *outMemory=(volatile int *)(ZBT_MEM +1);
#endif

static void j2k_write_eoc(opj_cio_t *cio);
static unsigned int cio_write(opj_cio_t *cio, unsigned int v, int n);
static int cio_tell(opj_cio_t *cio);
static void cio_skip(opj_cio_t *cio, int n);
static void cio_seek(opj_cio_t *cio, int pos);
static void j2k_write_soc(opj_cio_t *cio);
static void j2k_write_siz(opj_cio_t *cio,const infos_t * info);
static void j2k_write_cod(opj_cio_t *cio,const infos_t * info);
static void j2k_write_cox(opj_cio_t *cio, int compno,const infos_t * info);
static void j2k_write_qcd(opj_cio_t *cio,const infos_t * info);
static void j2k_write_qcx(opj_cio_t *cio, int compno,const infos_t * info);
static void j2k_write_coc(opj_cio_t *cio, int compno,const infos_t * info);
static void j2k_write_qcc(opj_cio_t *cio, int compno,const infos_t * info);
static void j2k_write_sot(opj_cio_t *cio,const packet_t * packets_in);
static int int_min(int a, int b);
static int int_max(int a, int b);
static int int_floorlog2(int a);
static void tcd_init_encode(infos_t * info_out, int num);
static void dwt_calc_explicit_stepsizes(int prec,infos_t * info);
static void dwt_encode_stepsize(int stepsize, int numbps,infos_t * info,int bandno);
static void j2k_setup_encoder(infos_t * info_out);

#endif /*VIDEOOUT_H_*/
