/*******************************************************************\

Copyright (c) 2008 Leiden University (LERC group at LIACS).
All rights reserved.

The use and distribution terms for this software are covered by the 
Common Public License 1.0 (http://opensource.org/licenses/cpl1.0.txt)
which can be found in the file LICENSE at the root of this distribution.
By using this software in any fashion, you are agreeing to be bound by 
the terms of this license.

You must not remove this notice, or any other, from this software.

\*******************************************************************/

#ifndef TYPES_H_
#define TYPES_H_

#if !defined(true)
#define true	1
#endif
#if !defined(false)
#define	false	0
#endif

//Number of tiles
//Horizontal
#define HTILES	8
//Vertical
#define VTILES	8

// Name of the image
#define FILE_NAME "baboon_128x128"
#define MAX_NAME  30
// Size of the image
#define WIDTH_IMAGE	128
#define HEIGHT_IMAGE 	128
// Size of the tile
#define WIDTH_TILE 	WIDTH_IMAGE / HTILES
#define HEIGHT_TILE HEIGHT_IMAGE / VTILES

//To define if DO NOT want to improve tiles with 'char's instead of 'int's
//#define INTEGER_TILES

// Number of components
#define NUM_COMPS 	3
// Number of resolutions
#define NUM_RES		1
// Number of bands per resolution
// (3 if more than one resolution otherwise 1)
#define NUM_BANDS	1
// Number of precints
#define NUM_PRECS 	1
// Number of codeblocks
#define NUM_CODEBS 	1

// If NUM_RES_MAX changes, change also the filling of RES_WIDTH & RES_HEIGTH in DWT.h
#define NUM_RES_MAX 4
#define RES_OFFSET NUM_RES_MAX - NUM_RES

#define MAX_COMP_SIZE 0
// Offset of the image
#define OFFSET_X	0
#define OFFSET_Y	0
#define SUBSAMPLING_DX	1
#define SUBSAMPLING_DY  1
#define TP_ON		0
// Progesion Order
#define PROG_ORDER	LRCP
#define NUM_POCS	0
#define CSTY		0
// Precission of the colour
#define PRECISION	8
// Depth of the colour
#define DEPTH		8
// 1: if signed value for the colour
#define SIGN		0
#define MAX_COMP_SIZE 0
#define CUR_TP_NUM  0
#define TP_POS 		0
#define CUR_PINO	0
#define MCT			1
#define MODE		0
#define QNTSTY		J2K_CCP_QNTSTY_NOQNT
#define NUMGBITS	2
#define CUR_TOTNUM_TP 1
#define ROI_COMPNO	-1
#define RES_SPEC	0
#define QMFBID	1
// Codeblock size (MA-a: if problems with single tile look HERE)
#define CBLKW	WIDTH_TILE
#define CBLKH	HEIGHT_TILE

// Maximal number of passes for a codeblock
#define MAX_PASSES	22
#define IMAGE_SIZE WIDTH_IMAGE * HEIGHT_IMAGE * NUM_COMPS

// Tile structure
typedef struct tile_t {
#ifdef INTEGER_TILES
	int comp[NUM_COMPS][WIDTH_TILE*HEIGHT_TILE];
#else 
	char comp[NUM_COMPS][WIDTH_TILE*HEIGHT_TILE];
#endif
	unsigned char num;
} tile_t;

typedef struct tile_c_t {
#ifdef INTEGER_TILES
	int comp[WIDTH_TILE*HEIGHT_TILE];
#else 
	char comp[WIDTH_TILE*HEIGHT_TILE];
#endif
	unsigned char num;
	unsigned char numcomp;
} tile_c_t;

// Codeblock structure
typedef struct codeblock_t {
	unsigned char data[NUM_RES][NUM_BANDS][NUM_PRECS][NUM_CODEBS][CBLKW*CBLKH]; // Should be CBLKW * CBLKH
	unsigned char num;
	unsigned char numcomp;
} codeblock_t;

// Passes structure
typedef struct pass_t{
  int term;
  int len;
  int rate;
  double distortiondec;
} pass_t;

typedef struct passes_t{
	pass_t passes[NUM_RES][NUM_BANDS][NUM_PRECS][NUM_CODEBS][MAX_PASSES];
	int passnum[NUM_RES][NUM_BANDS][NUM_PRECS][NUM_CODEBS];
	int numbps[NUM_RES][NUM_BANDS][NUM_PRECS][NUM_CODEBS];
	double distotile;
} passes_t;

typedef struct code_pas_t{
	passes_t p;
	codeblock_t c;
} code_pas_t;


typedef struct packet_t {
	unsigned char data[WIDTH_TILE*HEIGHT_TILE];
	int lenght;
	unsigned char num;
	unsigned char numcomp;
} packet_t;

#endif /*TYPES_H_*/

